import numpy as np
from math import sqrt

def noll_to_zernike(i):
    '''Get the Zernike index from a Noll index.

    Parameters
    ----------
    i : int
        The Noll index.

    Returns
    -------
    n : int
        The radial Zernike order.
    m : int
        The azimuthal Zernike order.
    '''
    n = int(sqrt(2 * i - 1) + 0.5) - 1
    if n % 2:
        m = 2 * int((2 * (i + 1) - n * (n + 1)) // 4) - 1
    else:
        m = 2 * int((2 * i + 1 - n * (n + 1)) // 4)
    return n, m * (-1)**(i % 2)

def ansi_to_zernike(i):
    '''Get the Zernike index from an ANSI index.

    Parameters
    ----------
    i : int
        The ANSI index.

    Returns
    -------
    n : int
        The radial Zernike order.
    m : int
        The azimuthal Zernike order.
    '''
    n = int((sqrt(8 * i + 1) - 1) / 2)
    m = 2 * i - n * (n + 2)
    return (n, m)

def zernike_to_ansi(n, m):
    '''Get the ANSI index for a pair of Zernike indices.

    Parameters
    ----------
    n : int
        The radial Zernike order.
    m : int
        The azimuthal Zernike order.

    Returns
    -------
    int
        The ANSI index.
    '''
    return (m + n * n) // 2 + n

def zernike_to_noll(n, m):
    '''Get the Noll index for a pair of Zernike indices.

    Parameters
    ----------
    n : int
        The radial Zernike order.
    m : int
        The azimuthal Zernike order.

    Returns
    -------
    int
        The Noll index.
    '''
    i = int(((n + 0.5)**2 + 1) / 2) + 1
    Nn = (n + 1) * (n + 2) // 2 + 1

    # Brute force search
    for j in range(i, i + Nn):
        nn, mm = noll_to_zernike(j)
        if nn == n and mm == m:
            return j
    raise ValueError('Could not find noll index for (%d,%d)' % n, m)

def zernike_radial(n, m, r, cache=None):
    '''The radial component of a Zernike polynomial.

    We use the q-recursive method, which uses recurrence relations to calculate the radial
    Zernike polynomials without using factorials. A description of the method can be found
    in [Chong2003]_. Additionally, this function optionally caches results of previous calls.

    .. [Chong2003] Chong, C. W., Raveendran, P., & Mukundan, R. (2003). A comparative analysis of algorithms
        for fast computation of Zernike moments. Pattern Recognition, 36(3), 731-742.

    Parameters
    ----------
    n : int
        The radial Zernike order.
    m : int
        The azimuthal Zernike order.
    r : array_like
        The (normalized) radial coordinates on which to calculate the polynomial.
    cache : dictionary or None
        A dictionary containing previously calculated Zernike modes on the same grid.
        This function is for speedup only, and therefore the cache is expected to be
        valid. You can reuse the cache for future calculations on the same exact grid.
        The given dictionary is updated with the current calculation.

    Returns
    -------
    array_like
        The radial component of the evaluated Zernike polynomial.
    '''
    m = abs(m)

    if cache is not None:
        if ('rad', n, m) in cache:
            return cache[('rad', n, m)]

    if n == m:
        res = r**n
    elif (n - m) == 2:
        z1 = zernike_radial(n, n, r, cache)
        z2 = zernike_radial(n - 2, n - 2, r, cache)

        res = n * z1 - (n - 1) * z2
    else:
        p = n
        q = m + 4

        h3 = -4 * (q - 2) * (q - 3) / float((p + q - 2) * (p - q + 4))
        h2 = h3 * (p + q) * (p - q + 2) / float(4 * (q - 1)) + (q - 2)
        h1 = q * (q - 1) / 2.0 - q * h2 + h3 * (p + q + 2) * (p - q) / 8.0

        r2 = zernike_radial(2, 2, r, cache)
        res = h1 * zernike_radial(p, q, r, cache) + (h2 + h3 / r2) * zernike_radial(n, q - 2, r, cache)

    if cache is not None:
        cache[('rad', n, m)] = res

    return res

def zernike_azimuthal(m, theta, cache=None):
    '''The azimuthal component of a Zernike polynomial.

    This function optionally caches results of previous calls.

    Parameters
    ----------
    m : int
        The azimuthal Zernike order.
    theta : array_like
        The azimuthal coordinates on which to calculate the polynomial.
    cache : dictionary or None
        A dictionary containing previously calculated Zernike modes on the same grid.
        This function is for speedup only, and therefore the cache is expected to be
        valid. You can reuse the cache for future calculations on the same exact grid.
        The given dictionary is updated with the current calculation.

    Returns
    -------
    array_like
        The azimuthal component of the evaluated Zernike polynomial.
    '''
    if cache is not None:
        if ('azim', m) in cache:
            return cache[('azim', m)]

    if m < 0:
        res = sqrt(2) * np.sin(-m * theta)
    elif m == 0:
        return 1
    else:
        res = sqrt(2) * np.cos(m * theta)

    if cache is not None:
        cache[('azim', m)] = res

    return res

def zernike(n, m, D=1, grid=None, radial_cutoff=True, cache=None):
    '''Evaluate the Zernike polynomial on a grid.

    Parameters
    ----------
    n : int
        The radial Zernike order.
    m : int
        The azimuthal Zernike order.
    D : scalar
        The diameter of the Zernike polynomial.
    grid : Grid
        The grid on which to evaluate the Zernike polynomial. If this is None,
        a Field generator will be returned.
    radial_cutoff : boolean
        Whether to apply a circular aperture to cutoff the modes.
    cache : dictionary or None
        A dictionary containing previously calculated Zernike modes on the same grid.
        This function is for speedup only, and therefore the cache is expected to be
        valid. You can reuse the cache for future calculations on the same exact grid.
        The given dictionary is updated with the current calculation.

    Returns
    -------
    Field or Field generator
        The evaluated Zernike polynomial. If `grid` is None, a Field generator is returned,
        which evaluates the Zernike polynomial on the supplied grid.
    '''
    from ..field import Field

    if grid is None:
        return lambda grid: zernike(n, m, D, grid, radial_cutoff, cache)

    if grid.is_separated and grid.is_('polar'):
        R, Theta = grid.separated_coords
        z_r = zernike_radial(n, m, 2 * R / D, cache)
        if radial_cutoff:
            z_r *= (2 * R) < D
        z = sqrt(n + 1) * np.outer(zernike_azimuthal(m, Theta, cache), z_r).flatten()
    else:
        r, theta = grid.as_('polar').coords
        z = sqrt(n + 1) * zernike_azimuthal(m, theta, cache) * zernike_radial(n, m, 2 * r / D, cache)
        if radial_cutoff:
            z *= (2 * r) < D

    return Field(z, grid)

def zernike_ansi(i, D=1, grid=None, radial_cutoff=True, cache=None):
    '''Evaluate the Zernike polynomial on a grid using an ANSI index.

    Parameters
    ----------
    i : int
        The ANSI index.
    D : scalar
        The diameter of the Zernike polynomial.
    grid : Grid or None
        The grid on which to evaluate the Zernike polynomial. If this is None,
        a Field generator will be returned.
    radial_cutoff : boolean
        Whether to apply a circular aperture to cutoff the modes.
    cache : dictionary or None
        A dictionary containing previously calculated Zernike modes on the same grid.
        This function is for speedup only, and therefore the cache is expected to be
        valid. You can reuse the cache for future calculations on the same exact grid.
        The given dictionary is updated with the current calculation.

    Returns
    -------
    Field or Field generator
        The evaluated Zernike polynomial. If `grid` is None, a Field generator is returned,
        which evaluates the Zernike polynomial on the supplied grid.
    '''
    n, m = ansi_to_zernike(i)
    return zernike(n, m, D, grid, radial_cutoff, cache)

def zernike_noll(i, D=1, grid=None, radial_cutoff=True, cache=None):
    '''Evaluate the Zernike polynomial on a grid using a Noll index.

    Parameters
    ----------
    i : int
        The Noll index.
    D : scalar
        The diameter of the Zernike polynomial.
    grid : Grid or None
        The grid on which to evaluate the Zernike polynomial. If this is None,
        a Field generator will be returned.
    radial_cutoff : boolean
        Whether to apply a circular aperture to cutoff the modes.
    cache : dictionary or None
        A dictionary containing previously calculated Zernike modes on the same grid.
        This function is for speedup only, and therefore the cache is expected to be
        valid. You can reuse the cache for future calculations on the same exact grid.
        The given dictionary is updated with the current calculation.

    Returns
    -------
    Field or Field generator
        The evaluated Zernike polynomial. If `grid` is None, a Field generator is returned,
        which evaluates the Zernike polynomial on the supplied grid.
    '''
    n, m = noll_to_zernike(i)
    return zernike(n, m, D, grid, radial_cutoff, cache)

def make_zernike_basis(num_modes, D, grid, starting_mode=1, ansi=False, radial_cutoff=True, use_cache=True):
    '''Make a ModeBasis of Zernike polynomials.

    Parameters
    ----------
    num_modes : int
        The number of Zernike polynomials to generate.
    D : scalar
        The diameter of the Zernike polynomial.
    grid : Grid or None
        The grid on which to evaluate the Zernike polynomials. If this is None,
        a list of Field generators will be returned.
    starting_mode : int
        The first mode to evaluate.
    ansi : boolean
        If this is True, the modes will be indexed using ANSI indices. Otherwise, a Noll
        indexing scheme is used.
    radial_cutoff : boolean
        Whether to apply a circular aperture to cutoff the modes.
    use_cache : boolean
        Whether to use a cache while calculating the modes. A cache uses memory, so turn it
        off when you are limited on memory.

    Returns
    -------
    ModeBasis or list of Field generators
        The evaluated mode basis of Zernike polynomials, or a list of Field generators for
        each of the Zernike polynomials.
    '''
    from .mode_basis import ModeBasis
    f = zernike_ansi if ansi else zernike_noll

    if grid is None:
        polar_grid = None
    else:
        polar_grid = grid.as_('polar')

    if use_cache:
        cache = {}
    else:
        cache = None

    modes = [f(i, D, polar_grid, radial_cutoff, cache) for i in range(starting_mode, starting_mode + num_modes)]

    if grid is None:
        return modes
    else:
        return ModeBasis(modes, grid)



__all__ = [
    'WavefrontSensorOptics',
    'WavefrontSensorEstimator',
    'make_odwfs_amplitude_filter',
    'make_polarization_odwfs_amplitude_filter',
    'optical_differentiation_surface',
    'OpticalDifferentiationWavefrontSensorOptics',
    'OpticalDifferentiationWavefrontSensorEstimator',
    'ModulatedPyramidWavefrontSensorOptics',
    'PyramidWavefrontSensorOptics',
    'PyramidWavefrontSensorEstimator',
    'ShackHartmannWavefrontSensorOptics',
    'SquareShackHartmannWavefrontSensorOptics',
    'ShackHartmannWavefrontSensorEstimator',
    'ZernikeWavefrontSensorOptics',
    'ZernikeWavefrontSensorEstimator',
    'VectorZernikeWavefrontSensorOptics'
]

from .optical_differentiation_wavefront_sensor import *
from .pyramid import *
from .shack_hartmann import *
from .wavefront_sensor import *
from .zernike_wavefront_sensor import *



from ..field import evaluate_supersampled
import numpy as np
from scipy.sparse import csr_matrix

def make_gaussian_pokes(grid, mu, sigma, cutoff=5):
    '''Make a basis of Gaussians.

    Parameters
    ----------
    grid : Grid or None
        The grid on which to calculate the mode basis. If this is None, a list of Field
        generators will be returned instead of a mode basis.
    mu : Grid
        The centers for each of the Gaussians.
    sigma : ndarray or scalar
        The standard deviation of each of the Gaussians. If this is a scalar,
        this value will be used for all Gaussians.
    cutoff : scalar or None
        The factor of sigma beyond which the Gaussian will be set to zero. The ModeBasis will
        be sparse to reduce memory usage. If the cutoff is None, there will be no cutoff, and
        the returned ModeBasis will be dense.

    Returns
    -------
    ModeBasis or list of Field generators.
        The sparse mode basis. If cutoff is None, a dense mode basis will be returned.
    '''
    sigma = np.ones(mu.size) * sigma

    def poke(m, s):
        def eval_func(func_grid):
            if func_grid.is_('cartesian'):
                r2 = (func_grid.x - m[0])**2 + (func_grid.y - m[1])**2
            else:
                r2 = func_grid.shifted(-m).as_('polar').r**2

            res = np.exp(-0.5 * r2 / s**2)

            if cutoff is not None:
                res -= np.exp(-0.5 * cutoff**2)
                res[r2 > (cutoff * s)**2] = 0

                res = csr_matrix(res)
                res.eliminate_zeros()

            return res
        return eval_func

    pokes = [poke(m, s) for m, s in zip(mu.points, sigma)]

    if grid is None:
        return pokes
    else:
        return evaluate_supersampled(pokes, grid, 1, make_sparse=True)



__all__ = [
    'get_strehl_from_focal',
    'get_strehl_from_pupil',
    'get_mean_intensity_in_roi',
    'get_mean_raw_contrast',
    'binned_profile',
    'azimutal_profile',
    'radial_profile'
]

from .contrast import *
from .profile import *



__all__ = [
    'FrameWriter',
    'GifWriter',
    'FFMpegWriter',
    'set_color_scheme',
    'colors',
    'errorfill',
    'imshow_field',
    'imsave_field',
    'contour_field',
    'contourf_field',
    'complex_field_to_rgb',
    'imshow_psf',
    'imshow_pupil_phase'
]

from .animation import *
from .color_scheme import *
from . import colors
from .error_bars import *
from .field import *
from .util import *



import numpy as np

from .wavefront import Wavefront, jones_to_mueller
from .optical_element import OpticalElement, make_agnostic_forward, make_agnostic_backward, AgnosticOpticalElement
from ..field import Field, field_dot, field_conjugate_transpose

def rotation_matrix(angle):
    '''Two dimensional rotation matrix.

    Parameters
    ----------
    angle : scaler
        rotation angle in radians

    Returns
    -------
    ndarray
        The rotation matrix.
    '''
    return np.array([[np.cos(angle), np.sin(angle)], [-np.sin(angle), np.cos(angle)]])

class JonesMatrixOpticalElement(AgnosticOpticalElement):
    '''A general Jones Matrix.

    Parameters
    ----------
    jones_matrix : matrix or Field
        The Jones matrix describing the optical element.
    '''
    def __init__(self, jones_matrix):
        self.jones_matrix = jones_matrix

        AgnosticOpticalElement.__init__(self, True, True)

    def make_instance(self, instance_data, input_grid, output_grid, wavelength):
        instance_data.jones_matrix = self.evaluate_parameter(self.jones_matrix, input_grid, output_grid, wavelength)

    def get_input_grid(self, output_grid, wavelength):
        return output_grid

    def get_output_grid(self, input_grid, wavelength):
        return input_grid

    @make_agnostic_forward
    def forward(self, instance_data, wavefront):
        '''Propgate the wavefront through the Jones matrix.

        Parameters
        ----------
        wavefront : Wavefront
            The wavefront to propagate.

        Returns
        -------
        Wavefront
            The propagated wavefront.
        '''
        if wavefront.is_scalar:
            # We generate an unpolarized wavefront
            if hasattr(instance_data.jones_matrix, 'grid'):
                electric_field = instance_data.jones_matrix * wavefront.electric_field
            else:
                electric_field = instance_data.jones_matrix[..., np.newaxis] * wavefront.electric_field

            return Wavefront(electric_field, wavelength=wavefront.wavelength, input_stokes_vector=[1, 0, 0, 0])
        else:
            wf = wavefront.copy()
            wf.electric_field = field_dot(instance_data.jones_matrix, wf.electric_field)

            return wf

    @make_agnostic_backward
    def backward(self, instance_data, wavefront):
        '''Propagate the wavefront backwards through the Jones matrix.

        Parameters
        ----------
        wavefront : Wavefront
            The wavefront to propagate.

        Returns
        -------
        Wavefront
            The propagated wavefront.
        '''
        if wavefront.is_scalar:
            # We generate an unpolarized wavefront
            if hasattr(instance_data.jones_matrix, 'grid'):
                electric_field = field_conjugate_transpose(instance_data.jones_matrix) * wavefront.electric_field
            else:
                electric_field = field_conjugate_transpose(instance_data.jones_matrix)[..., np.newaxis] * wavefront.electric_field

            return Wavefront(electric_field, wavelength=wavefront.wavelength, input_stokes_vector=[1, 0, 0, 0])
        else:
            wf = wavefront.copy()
            wf.electric_field = field_dot(field_conjugate_transpose(instance_data.jones_matrix), wf.electric_field)

            return wf

    @property
    def mueller_matrix(self):
        '''Returns the Mueller matrix corresponding to the Jones matrix.
        '''
        return self.construct_function(jones_to_mueller, self.jones_matrix)

    def __mul__(self, b):
        '''Multiply the Jones matrix on the right side with the Jones matrix m.

        Parameters
        ----------
        m : JonesMatrixOpticalElement, Field or matrix
            The Jones matrix with which to multiply.
        '''
        if hasattr(b, 'jones_matrix'):
            jones_matrix = self.construct_function(field_dot, self.jones_matrix, b.jones_matrix)
        else:
            jones_matrix = self.construct_function(field_dot, self.jones_matrix, b)

        return JonesMatrixOpticalElement(jones_matrix)

class PhaseRetarder(JonesMatrixOpticalElement):
    '''A general phase retarder.

    Parameters
    ----------
    phase_retardation : scalar or Field
        The relative phase retardation induced between the fast and slow axis.
    fast_axis_orientation : scalar or Field
        The angle of the fast axis with respect to the x-axis in radians.
    circularity : scalar or Field
        The circularity of the phase retarder.
    '''
    def __init__(self, phase_retardation, fast_axis_orientation, circularity):
        self._phase_retardation = phase_retardation
        self._fast_axis_orientation = fast_axis_orientation
        self._circularity = circularity

        JonesMatrixOpticalElement.__init__(self, self.jones_matrix)

    @property
    def jones_matrix(self):
        def jones(phase_retardation, fast_axis_orientation, circularity):
            phi_plus = np.exp(1j * phase_retardation / 2)
            phi_minus = np.exp(-1j * phase_retardation / 2)

            # calculating the individual components
            j11 = phi_plus * np.cos(fast_axis_orientation)**2 + phi_minus * np.sin(fast_axis_orientation)**2
            j12 = (phi_plus - phi_minus) * np.exp(-1j * circularity) * np.cos(fast_axis_orientation) * np.sin(fast_axis_orientation)
            j21 = (phi_plus - phi_minus) * np.exp(1j * circularity) * np.cos(fast_axis_orientation) * np.sin(fast_axis_orientation)
            j22 = phi_plus * np.sin(fast_axis_orientation)**2 + phi_minus * np.cos(fast_axis_orientation)**2

            # constructing the Jones matrix.
            jones_matrix = np.array([[j11, j12], [j21, j22]])

            if hasattr(j11, 'grid'):
                jones_matrix = Field(jones_matrix, j11.grid)

            return jones_matrix

        return self.construct_function(jones, self.phase_retardation, self.fast_axis_orientation, self.circularity)

    @jones_matrix.setter
    def jones_matrix(self, b):
        pass

    @property
    def phase_retardation(self):
        return self._phase_retardation

    @phase_retardation.setter
    def phase_retardation(self, phase_retardation):
        self._phase_retardation = phase_retardation

        self.clear_cache()

    @property
    def fast_axis_orientation(self):
        return self._fast_axis_orientation

    @fast_axis_orientation.setter
    def fast_axis_orientation(self, fast_axis_orientation):
        self._fast_axis_orientation = fast_axis_orientation

        self.clear_cache()

    @property
    def circularity(self):
        return self._circularity

    @circularity.setter
    def circularity(self, circularity):
        self._circularity = circularity

        self.clear_cache()

class LinearRetarder(PhaseRetarder):
    '''A general linear retarder.

    Parameters
    ----------
    phase_retardation : scalar or Field
        The relative phase retardation induced between the fast and slow axis.
    fast_axis_orientation : scalar or Field
        The angle of the fast axis with respect to the x-axis in radians.
    '''
    def __init__(self, phase_retardation, fast_axis_orientation):
        PhaseRetarder.__init__(self, phase_retardation, fast_axis_orientation, 0)

class CircularRetarder(PhaseRetarder):
    '''A general circular retarder.

    Parameters
    ----------
    phase_retardation : scalar or Field
        The relative phase retardation induced between the fast and slow axis.
    fast_axis_orientation : scalar or Field
        The angle of the fast axis with respect to the x-axis in radians.
    '''
    def __init__(self, phase_retardation):
        PhaseRetarder.__init__(self, phase_retardation, np.pi / 4, np.pi / 2)

class QuarterWavePlate(LinearRetarder):
    '''A quarter-wave plate.

    Parameters
    ----------
    fast_axis_orientation : scalar or Field
        The angle of the fast axis with respect to the x-axis in radians.
    '''
    def __init__(self, fast_axis_orientation):
        LinearRetarder.__init__(self, np.pi / 2, fast_axis_orientation)

class HalfWavePlate(LinearRetarder):
    '''A half-wave plate.

    Parameters
    ----------
    fast_axis_orientation : scalar or Field
        The angle of the fast axis with respect to the x-axis in radians.
    '''
    def __init__(self, fast_axis_orientation):
        LinearRetarder.__init__(self, np.pi, fast_axis_orientation)

class GeometricPhaseElement(LinearRetarder):
    '''A general geometric phase element.

    Parameters
    ----------
    phase_pattern : Field or array_like
        The phase pattern in radians.
    leakage : scalar
        The relative leakage strength (0 = no leakage, 1 = maximum leakage)
    retardance_offset : scalar
        The retardance offset from half wave in radians. This will result in leakage.
    '''
    def __init__(self, phase_pattern, leakage=None, retardance_offset=0):
        if leakage is not None:
            # Testing if the leakage is valid.
            if leakage < 0 or leakage > 1:
                raise ValueError('Leakage must be between 0 and 1.')

            # Calculating the required retardance offset to get the leakage.
            retardance_offset = 2 * np.arcsin(np.sqrt(leakage))

        LinearRetarder.__init__(self, np.pi - retardance_offset, phase_pattern / 2)

class LinearPolarizer(JonesMatrixOpticalElement):
    '''A linear polarizer.

    Parameters
    ----------
    polarization_angle : scalar or Field
        The polarization angle of the polarizer. Light with this angle is transmitted.
    '''
    def __init__(self, polarization_angle):
        self.polarization_angle = polarization_angle

        JonesMatrixOpticalElement.__init__(self, self.jones_matrix)

    @property
    def jones_matrix(self):
        def jones(angle):
            c = np.cos(angle)
            s = np.sin(angle)

            return np.array([[c**2, c * s], [c * s, s**2]])

        return self.construct_function(jones, self.polarization_angle)

    @jones_matrix.setter
    def jones_matrix(self, b):
        pass

    @property
    def polarization_angle(self):
        '''The angle of polarization of the linear polarizer.
        '''
        return self._polarization_angle

    @polarization_angle.setter
    def polarization_angle(self, angle):
        self._polarization_angle = angle

        self.clear_cache()

class LinearPolarizingBeamSplitter(OpticalElement):
    ''' A linear polarizing beam splitter that accepts one wavefront and returns two orthogonally linearly polarized wavefronts.

    The first of the returned wavefront will have the polarization state set by the polarization angle. The second wavefront
    will be perpendicular to the first.

    Parameters
    ----------
    polarization_angle : scalar or Field
        The polarization angle of the polarizer.
    '''
    def __init__(self, polarization_angle, wavelength=1):
        self.polarization_angle = polarization_angle

    @property
    def polarization_angle(self):
        '''The angle of polarization of the linear polarizer.
        '''
        return self._polarization_angle

    @polarization_angle.setter
    def polarization_angle(self, polarization_angle):
        self._polarization_angle = polarization_angle

        # Calculating two Jones matrices for the two ports.
        self.polarizer_port_1 = LinearPolarizer(polarization_angle)
        self.polarizer_port_2 = LinearPolarizer(polarization_angle + np.pi / 2)

    def forward(self, wavefront):
        '''Propgate the wavefront through the PBS.

        Parameters
        ----------
        wavefront : Wavefront
            The wavefront to propagate.

        Returns
        -------
        wf_1 : Wavefront
            The wavefront propagated through a polarizer under the polarization angle.

        wf_2 : Wavefront
            The propagated wavefront through a polarizer perpendicular to the first one.
        '''

        wf_1 = self.polarizer_port_1.forward(wavefront)
        wf_2 = self.polarizer_port_2.forward(wavefront)

        return wf_1, wf_2

    def backward(self, wavefront):
        '''Propagate the wavefront backwards through the PBS.

        Not possible, will raise error.
        '''
        raise RuntimeError('Backward propagation through PolarizingBeamSplitter not possible.')

    @property
    def mueller_matrices(self):
        '''Returns the Mueller matrices of the two Jones matrices.
        '''
        return jones_to_mueller(self.polarizer_port_1.jones_matrix), jones_to_mueller(self.polarizer_port_2.jones_matrix)

class CircularPolarizingBeamSplitter(OpticalElement):
    ''' A circular polarizing beam splitter that accepts one wavefront and returns two linearly polarized wavefronts.

    The circular polarizing beam splitter is a combination of a quarter-wave plate and a polarizing beam splitter. The
    quarter-wave plate rotation angle is 45 degrees and the polarizing beam-splitter rotation angle is zero degrees.
    Therefore, the first of the returned wavefronts will have a linear polarization state with an electric field equal
    to the amount of left-circular polarization in the incoming wavefront. The second wavefront will be perpendicular
    to the first and with electric field equal to the amount of right-circular polarization in the incoming wavefront.

    Parameters
    ----------
    polarization_angle : scalar or Field
        The polarization angle of the polarizer.
    '''
    def __init__(self, wavelength=1):
        self.polarization_angle = 0
        self.quarter_wave_angle = np.pi / 4
        self.quarter_wave_plate = QuarterWavePlate(self.quarter_wave_angle)
        self.linear_polarizing_beam_splitter = LinearPolarizingBeamSplitter(self.polarization_angle)

    def forward(self, wavefront):
        '''Propgate the wavefront through the CBS.

        Parameters
        ----------
        wavefront : Wavefront
            The wavefront to propagate.

        Returns
        -------
        wf_1 : Wavefront
            The wavefront propagated through a quarter-wave plate at 45 degrees and a polarizer 0 degrees.
        wf_2 : Wavefront
            The wavefront propagated through a quarter-wave plate at 45 degrees and a polarizer 90 degrees.
        '''
        wf_1, wf_2 = self.linear_polarizing_beam_splitter.forward(self.quarter_wave_plate.forward(wavefront))
        return wf_1, wf_2

    def backward(self, wavefront):
        '''Propagate the wavefront backwards through the CBS.

        Not possible, will raise error.
        '''
        raise RuntimeError('Backward propagation through PolarizingBeamSplitter not possible.')

    @property
    def mueller_matrices(self):
        '''Returns the Mueller matrices of the two Jones matrices.
        '''
        qwp_mueller = self.quarter_wave_plate.mueller_matrix
        lin_pol_bs_mueller_1, lin_pol_bs_mueller_2 = self.linear_polarizing_beam_splitter.mueller_matrices

        return np.dot(lin_pol_bs_mueller_1, qwp_mueller), np.dot(lin_pol_bs_mueller_2, qwp_mueller)



__all__ = [
    'ModeBasis',
    'make_gaussian_hermite_basis',
    'gaussian_hermite_index',
    'gaussian_hermite',
    'index_to_hermite',
    'gaussian_laguerre',
    'make_gaussian_laguerre_basis',
    'make_gaussian_pokes',
    'make_sine_basis',
    'make_cosine_basis',
    'make_fourier_basis',
    'make_complex_fourier_basis',
    'make_lp_modes',
    'make_LP_modes',
    'make_zernike_basis',
    'zernike',
    'noll_to_zernike',
    'zernike_to_noll',
    'ansi_to_zernike',
    'zernike_to_ansi',
    'disk_harmonic',
    'disk_harmonic_energy',
    'make_disk_harmonic_basis'
]

from .disk_harmonic import *
from .fourier import *
from .gaussian_hermite import *
from .gaussian_laguerre import *
from .gaussian_pokes import *
from .karhunen_loeve import *
from .mode_basis import *
from .lp_fiber_modes import *
from .zernike import *



from hcipy import *
from matplotlib import pyplot as plt
import numpy as np

def test_pupil_plane_wavefront_sensors():
    # Create the input grid
    D = 1
    pupil_grid = make_pupil_grid(128, D)
    aperture = make_circular_aperture(D)

    # Make zernike basis
    num_zernike = 3
    zernike_basis = make_zernike_basis(num_zernike, D, pupil_grid, 2)

    # Create wavefront sensors
    pyramid_wfs = PyramidWavefrontSensorOptics(pupil_grid)
    zernike_wfs = ZernikeWavefrontSensorOptics(pupil_grid, q=32, num_airy=20)
    god_wfs = gODWavefrontSensorOptics(40, 1/3, pupil_grid, pupil_separation=1.5, q=16, num_airy=40)
    pod_wfs = PolgODWavefrontSensorOptics(40, 1/3, pupil_grid, pupil_separation=1.5, q=16, num_airy=40)

    wavefront_sensors = (pyramid_wfs, zernike_wfs, god_wfs, pod_wfs)
    num_wfs = len(wavefront_sensors)

    # Create reference images
    wf = Wavefront(aperture(pupil_grid))
    wf.total_power = 1
    refs = [wfs.forward(wf).intensity for wfs in wavefront_sensors]
    eff = [wfs.forward(wf).total_power for wfs in wavefront_sensors]
    print("Total efficiency : ", eff)

    for i in range(num_wfs):
        plt.subplot(2,2,i+1)
        imshow_field(refs[i], wavefront_sensors[i].output_grid)
    plt.show()

    for i, zern in enumerate(zernike_basis):
        # create input
        wf = Wavefront(aperture(pupil_grid) * np.exp(1j * 0.01 * zern))
        wf.total_power = 1

        # Measure the response of all wavefront sensors
        for j, wfs in enumerate(wavefront_sensors):
            wf_out = wfs.forward(wf)
            print(wf_out.total_power)
            plt.subplot(num_wfs, num_zernike, i+1 + num_zernike*j)
            imshow_field(wf_out.intensity-refs[j])

    plt.show()



if __name__ == "__main__":
    test_pupil_plane_wavefront_sensors()



import heapq

class DynamicOpticalSystem(object):
    '''A dynamically varying optical system.

    This can be used as a starting point for an implementation of an adaptive optics system. The optics
    are changed as specific moments in time, which can be scheduled by submitting a callback. The time
    points are bridged using the integrate function, which is written by the derived class. This
    function should integrate on each detector without reading out the detector. Detector readouts or
    DM changes should be implemented as callbacks.
    '''
    def __init__(self):
        self.callbacks = []
        self.t = 0
        self.callback_counter = 0

    def evolve_until(self, t):
        '''Evolve the optical system until time `t`.

        Parameters
        ----------
        t : scalar
            The point in time to which to simulate this optical system. Callbacks and integrations
            are handled along the way.
        '''
        if t < self.t:
            raise ValueError('Backwards evolution is not allowed.')

        end = False
        while not end:
            t_next, _, callback = self.callbacks[0]
            if t_next < t:
                integration_time = t_next - self.t
                heapq.heappop(self.callbacks)
            else:
                integration_time = t - self.t
                end = True

            # Avoid multiple expensive integrations if we have multiple callbacks
            # at the same time.
            if integration_time > 1e-6:
                self.integrate(integration_time)
                self.t += integration_time

            if not end:
                callback()

    def integrate(self, integration_time):
        '''Integrate the current optical system for a certain integration time.

        This function should be implemented by a user in a derived class.

        Parameters
        ----------
        integration_time : scalar
            The integration time with which to integrate the optical system.
        '''
        pass

    def add_callback(self, t, callback):
        '''Add a callback to the list of callbacks.

        This function can even be called during handling of a callback. This is
        especially useful when implementing periodic callbacks: a callback can
        reinsert itself at a later time at the end of handling of the callback.

        Parameters
        ----------
        t : scalar
            The time at which to call the callback.
        callback : function
            The function to call at time `t`. This function should have no arguments.
        '''
        # Callback counter is to avoid comparison of callback functions
        # if the times are equal, as comparison of functions is not allowed.
        heapq.heappush(self.callbacks, (t, self.callback_counter, callback))
        self.callback_counter += 1



__all__ = [
    'make_linear_interpolator_separated',
    'make_linear_interpolator_unstructured',
    'make_linear_interpolator',
    'make_nearest_interpolator_separated',
    'make_nearest_interpolator_unstructured',
    'make_nearest_interpolator'
]

from .linear import *
from .nearest import *



from .field import Field
import numpy as np
import string

def field_einsum(subscripts, *operands, **kwargs):
    '''Evaluates the Einstein summation convention on the operand fields.

    This function uses the same conventions as numpy.einsum(). The input
    subscript is multiplexed over each position in the grid. The grids of each of
    the input field operands don't have to match, but must have the same lengths.

    The subscripts must be written as you would for a single position in the grid.
    The function alters these subscripts to multiplex over the entire grid.

    .. caution::
        Some subscripts may yield no exception, even though they would fail for
        a single point in the grid. The output in these cases can not be trusted.

    Parameters
    ----------
    subscripts : str
        Specifies the subscripts for summation.
    operands : list of array_like or `Field`
        These are the arrays or fields for the operation.
    out : {ndarray, None}, optional
        If provided, the calculation is done into this array.
    dtype : {data-type, None}, optional
        If provided, forces the calculation to use the data type specified.
        Note that you may have to also give a more liberal `casting`
        parameter to allow the conversions. Default is None.
    order : {'C', 'F', 'A', 'K'}, optional
        Controls the memory layout of the output. 'C' means it should
        be C contiguous. 'F' means it should be Fortran contiguous,
        'A' means it should be 'F' if the inputs are all 'F', 'C' otherwise.
        'K' means it should be as close to the layout as the inputs as
        is possible, including arbitrarily permuted axes.
        Default is 'K'.
    casting : {'no', 'equiv', 'safe', 'same_kind', 'unsafe'}, optional
        Controls what kind of data casting may occur.  Setting this to
        'unsafe' is not recommended, as it can adversely affect accumulations.

            * 'no' means the data types should not be cast at all.
            * 'equiv' means only byte-order changes are allowed.
            * 'safe' means only casts which can preserve values are allowed.
            * 'same_kind' means only safe casts or casts within a kind,
                like float64 to float32, are allowed.
            * 'unsafe' means any data conversions may be done.

        Default is 'safe'.
    optimize : {False, True, 'greedy', 'optimal'}, optional
        Controls if intermediate optimization should occur. No optimization
        will occur if False and True will default to the 'greedy' algorithm.
        Also accepts an explicit contraction list from the ``np.einsum_path``
        function. See ``np.einsum_path`` for more details. Default is False.

    Returns
    -------
    Field
        The calculated Field based on the Einstein summation convention.

    Raises
    ------
    ValueError
        If all of the fields don't have    the same grid size. If the number of
        operands is not equal to the number of subscripts specified.
    '''
    is_field = [isinstance(o, Field) for o in operands]
    if not np.count_nonzero(is_field):
        return np.einsum(subscripts, *operands, **kwargs)

    field_sizes = [o.grid.size for i, o in enumerate(operands) if is_field[i]]
    element_sizes = [o.shape[-1] for i, o in enumerate(operands) if is_field[i]]

    if not np.allclose(field_sizes, field_sizes[0]) or not np.allclose(element_sizes, element_sizes[0]):
        raise ValueError('All fields must be the same size for a field_einsum().')

    # Decompose the subscript into input and output
    splitted_string = subscripts.split('->')
    if len(splitted_string) == 2:
        ss_input, ss_output = splitted_string
    else:
        ss_input = splitted_string[0]
        ss_output = ''

    # split the input operands in separate strings
    ss = ss_input.split(',')
    if len(ss) != len(operands):
        raise ValueError('Number of operands is not equal to number of indexing operands.')

    # Find an indexing letter that can be used for field dimension.
    unused_index = [a for a in string.ascii_lowercase if a not in subscripts][0]

    # Add the field dimension to the input field operands.
    ss = [s + unused_index if is_field[i] else s for i, s in enumerate(ss)]

    # Recombine all operands into the final subscripts
    if len(splitted_string) == 2:
        subscripts_new = ','.join(ss) + '->' + ss_output + unused_index
    else:
        subscripts_new = ','.join(ss)

    res = np.einsum(subscripts_new, *operands, **kwargs)
    grid = operands[np.flatnonzero(np.array(is_field))[0]].grid

    if 'out' in kwargs:
        kwargs['out'] = Field(res, grid)
    return Field(res, grid)

def field_dot(a, b, out=None):
    '''Perform a dot product of `a` and `b` multiplexed over the field dimension.

    Parameters
    ----------
    a : Field or array_like
        Left argument of the dot product.
    b : Field or array_like
        Right argument of the dot product.
    out : Field or array_like
        If provided, the calculation is done into this array.

    Returns
    -------
    Field
        The result of the dot product.
    '''
    # Find out if a or b are vectors or higher dimensional tensors
    if hasattr(a, 'tensor_order'):
        amat = a.tensor_order > 1
    elif np.isscalar(a):
        if out is None:
            return a * b
        else:
            return np.multiply(a, b, out)
    else:
        amat = a.ndim > 1

    if hasattr(b, 'tensor_order'):
        bmat = b.tensor_order > 1
    elif np.isscalar(b):
        if out is None:
            return a * b
        else:
            return np.multiply(a, b, out)
    else:
        bmat = b.ndim > 1

    # Select correct multiplication behaviour.
    if amat and bmat:
        subscripts = '...ij,...jk->...ik'
    elif amat and not bmat:
        subscripts = '...i,...i->...'
    elif not amat and bmat:
        subscripts = '...i,...ij->...j'
    elif not amat and not bmat:
        subscripts = '...i,...i->...'

    # Perform calculation and return.
    if out is None:
        return field_einsum(subscripts, a, b)
    else:
        return field_einsum(subscripts, a, b, out=out)

def field_trace(a, out=None):
    '''Take the trace of a tensor field.

    Parameters
    ----------
    a : Field
        The field of which to take the trace.
    out : Field or None
        If given, a location where the output is stored. This must have the correct shape.

    Returns
    -------
    Field
        The trace of the field `a`.
    '''
    if out is None:
        return field_einsum('ii', a)
    else:
        return field_einsum('ii', a, out=out)

def field_inverse_tikhonov(f, rcond=1e-15):
    '''Invert a tensor field of order 2 using Tikhonov regularization.

    Parameters
    ----------
    f : `Field` or ndarray
        The tensor field for which to calculate the inverses. The tensor order
        of this Field has to be 2. If it is not a field, a normal inverse with
        Tikhonov regularization will be returned.
    rcond : scalar
        The relative regularization parameter to use for the inversions.

    Returns
    -------
    Field or ndarray
        The resulting Field with tensor order 2.

    Raises
    ------
    ValueError
        If the tensor order of field `f` is not 2.
    '''
    from ..util import inverse_tikhonov

    if hasattr(f, 'grid'):
        if f.tensor_order != 2:
            raise ValueError("Field must be a tensor field of order 2 to be able to calculate inverses.")

        res = np.empty((f.tensor_shape[1], f.tensor_shape[0], f.grid.size))
        for i in range(f.grid.size):
            res[..., i] = inverse_tikhonov(f[..., i], rcond)
        return Field(res, f.grid)
    else:
        return inverse_tikhonov(f, rcond)

def field_inverse_truncated(f, rcond=1e-15):
    '''Invert a tensor field of order 2 using mode truncation.

    Parameters
    ----------
    f : `Field` or ndarray
        The tensor field for which to calculate the inverses. The tensor order
        of this Field has to be 2. If it is not a field, a normal truncated
        inverse will be returned.
    rcond : scalar
        The relative condition number of the highest-order mode that must
        be used for inversion.

    Returns
    -------
    Field or ndarray
        The resulting Field with tensor order 2.

    Raises
    ------
    ValueError
        If the tensor order of field `f` is not 2.
    '''
    from ..util import inverse_truncated

    if hasattr(f, 'grid'):
        if f.tensor_order != 2:
            raise ValueError("Field must be a tensor field of order 2 to be able to calculate inverses.")

        res = np.empty((f.tensor_shape[1], f.tensor_shape[0], f.grid.size))
        for i in range(f.grid.size):
            res[..., i] = inverse_truncated(f[..., i], rcond)
        return Field(res, f.grid)
    else:
        return inverse_truncated(f, rcond)

def field_inverse_truncated_modal(f, num_modes):
    '''Invert a tensor field of order 2 using mode truncation.

    Parameters
    ----------
    f : `Field` or ndarray
        The tensor field for which to calculate the inverses. The tensor order
        of this Field has to be 2. If it is not a field, a normal truncated
        inverse will be returned
    rcond : scalar
        The relative condition number of the highest-order mode that must
        be used for inversion.

    Returns
    -------
    Field or ndarray
        The resulting Field with tensor order 2.

    Raises
    ------
    ValueError
        If the tensor order of field `f` is not 2.
    '''
    from ..util import inverse_truncated_modal

    if hasattr(f, 'grid'):
        if f.tensor_order != 2:
            raise ValueError("Field must be a tensor field of order 2 to be able to calculate inverses.")

        res = np.empty((f.tensor_shape[1], f.tensor_shape[0], f.grid.size))
        for i in range(f.grid.size):
            res[..., i] = inverse_truncated_modal(np.asarray(f[..., i]), num_modes)
        return Field(res, f.grid)
    else:
        return inverse_truncated_modal(f, num_modes)

def field_inv(f):
    '''Compute the (multiplicative) inverse of a tensor field of order 2.

    Parameters
    ----------
    f : `Field`
        The tensor field for which to calculate the inverses. The tensor order
        of this field has to be 2.

    Returns
    -------
    Field
        The resulting Field with tensor order 2.

    Raises
    ------
    ValueError
        If the tensor order of field `f` is not 2.
    '''
    if hasattr(f, 'grid'):
        if f.tensor_order != 2:
            raise ValueError('Field must be a tensor field of order 2 to be able to compute inverses.')

        res = np.moveaxis(np.linalg.inv(np.moveaxis(f, -1, 0)), 0, -1)
        return Field(res, f.grid)
    else:
        return np.linalg.inv(f)

def field_svd(f, full_matrices=True, compute_uv=True):
    '''Calculate the singular value decomposition for a tensor field of order 2.

    Parameters
    ----------
    f : `Field`
        The tensor field for which to calculate the singular value decompositions.
    full_matrices : boolean
        If True, matrices in U and Vh will have shapes (M,M) and (N,N) respectively.
        Otherwise their shapes are (M,K), (K,N) respectively, where K=min(M,N).
    compute_uv : boolean
        Whether to compute matrices U and Vh in addition to the singular values.

    Returns
    -------
    U : `Field`
        The unitary matrices. Only returned if `compute_uv` is True.
    S : `Field`
        The singular values, sorted in descending order.
    Vh : `Field`
        The unitary matrices. Only returned if `compute_uv` is True.
    '''

    res = np.linalg.svd(np.moveaxis(f, -1, 0), full_matrices, compute_uv)

    if compute_uv:
        U, S, Vh = res
        U = Field(np.moveaxis(U, 0, -1), f.grid)
        Vh = Field(np.moveaxis(Vh, 0, -1), f.grid)
    else:
        S = res

    S = Field(np.moveaxis(S, 0, -1), f.grid)

    if compute_uv:
        return U, S, Vh
    else:
        return S

def field_conjugate_transpose(a):
    '''Performs the conjugate transpose of a rank 2 tensor field or two dimensional array.

    Parameters
    ----------
    a : Field or array
        The element to conjugate transpose

    Returns
    -------
    Field or array
        The conjugate transposed element
    '''

    # first we test if it's a field
    if hasattr(a, 'tensor_order'):
        # if its a field, it must have a rank of 2
        if a.tensor_order != 2:
            raise ValueError('Need a tensor field of rank 2.')

        return Field(np.swapaxes(a.conj(), 0, 1), a.grid)
    else:
        # if its an array, it must be two dimensional
        if len(a.shape) != 2:
            raise ValueError('Need a two dimensional array.')

        return np.swapaxes(a.conj(), 0, 1)

def field_transpose(a):
    '''Performs the transpose of a rank 2 tensor field or two dimensional array.

    Parameters
    ----------
    a : Field or array
        The element to transpose

    Returns
    -------
    Field or array
        The transposed field or array
    '''
    # first we test if it's a field
    if hasattr(a, 'tensor_order'):
        # if its a field, it must have a rank of 2
        if a.tensor_order != 2:
            raise ValueError('Need a tensor field of rank 2.')

        return Field(np.swapaxes(a, 0, 1), a.grid)
    else:
        # if its an array, it must be two dimensional
        if len(a.shape) != 2:
            raise ValueError('Need a two dimensional array.')

        return np.swapaxes(a, 0, 1)

def field_determinant(a):
    '''Calculates the determinant of a tensor field.

    Parameters
    ----------
    a : Field
        The field for which the determinant needs to be calculated

    Returns
    -------
    Field
        The field that contains the determinant on every spatial position
    '''
    if a.tensor_order == 1:
        raise ValueError('Only tensor fields of order 2 or higher have a determinant.')

    if a.tensor_order > 2:
        raise NotImplementedError()

    if not np.all(a.tensor_shape == a.tensor_shape[0]):
        raise ValueError('Need square matrix for determinant.')

    # First we need to swap the axes in order to use np.linalg.det
    Temp = np.swapaxes(a, 0, 2)

    return Field(np.linalg.det(Temp), a.grid)

def field_adjoint(a):
    '''Calculates the adjoint of a tensor field.

    Parameters
    ----------
    a : Field
        The field for which the adjoint needs to be calculated

    Returns
    -------
    Field
        The adjointed field
    '''
    if a.tensor_order != 2:
        raise ValueError('Only tensor fields of order 2 can be inverted.')

    # Calculating the determinant.
    determinant = field_determinant(a)

    if np.any(np.isclose(determinant, 0)):
        raise ValueError('Matrix is non-invertible due to zero determinant.')

    return Field(determinant[np.newaxis, np.newaxis, :] * field_inv(a), a.grid)

def field_cross(a, b):
    '''Calculates the cross product of two vector fields.

    Parameters
    ----------
    a : Field
        The first field of the cross product
    b : Field
        The second field of the cross product

    Returns
    -------
    Field
        The cross product field
    '''
    if a.tensor_order != 1 or b.tensor_order != 1:
        raise ValueError('Only tensor fields of order 1 can have a cross product.')

    if a.shape[0] != 3 or b.shape[0] != 3:
        raise ValueError('Vector needs to be of length 3 for cross product.')

    return Field(np.cross(a, b, axis=0), a.grid)

def field_kron(a, b):
    '''Calculate the Kronecker product of two fields.

    Parameters
    ----------
    a : tensor Field
        The first Field
    b : tensor Field
        The second Field

    Returns
    -------
    Field
        The resulting tensor field.
    '''
    is_a_field = hasattr(a, 'grid')
    is_b_field = hasattr(b, 'grid')

    is_output_field = is_a_field or is_b_field

    if not is_output_field:
        return np.kron(a, b)

    if is_a_field and is_b_field:
        if a.grid.size != b.grid.size:
            raise ValueError('Field sizes for a (%d) and b (%d) are not compatible.' % (a.grid.size, b.grid.size))
        grid = a.grid
    else:
        if is_a_field:
            grid = a.grid
        else:
            grid = b.grid

    if is_a_field:
        aa = a
    else:
        aa = a[..., np.newaxis]

    if is_b_field:
        bb = b
    else:
        bb = b[..., np.newaxis]

    output_tensor_shape = np.array(aa.shape[:-1]) * np.array(bb.shape[:-1])
    output_shape = np.concatenate((output_tensor_shape, [grid.size]))

    res = (aa[:, np.newaxis, :, np.newaxis, :] * bb[np.newaxis, :, np.newaxis, :, :]).reshape(output_shape)

    return Field(res, grid)

def make_field_operation(op):
    pass
'''
def make_field_operation(op):
    def inner(*args, **kwargs):
        # Determine which args are fields.
        is_field = [hasattr(arg, 'grid') for arg in args]

        if not np.any(is_field):
            return op(*args, **kwargs)

        # At least one argument has a field
        grid_size = np.flatnonzero(is_field)[0]].grid.size

        if len(args) == 1:
            # Only one argument; use loop comprehension
            res = np.array([op(args[0][...,i]) for i in range(grid_size)])

            return Field(, args[0].grid)

        # More than one argument operation.
        res = []
        for i in range(grid_size):
            a = tuple([args[j][...,i] if is_field[j] else args[j] for j in len(args)])
            res.append(op(*a, **kwargs))
        return Field(res, )
        '''



from scipy.interpolate import RegularGridInterpolator, NearestNDInterpolator
import numpy as np
from ..field import Field

def make_nearest_interpolator_separated(field, grid=None):
    '''Make a nearest interpolator for a field on a separated grid.

    Parameters
    ----------
    field : Field or ndarray
        The field to interpolate.
    grid : Grid or None
        The grid of the field. If it is given, the grid of `field` is replaced by this grid.

    Returns
    -------
    Field generator
        The interpolator, as a Field generator. The grid on which this field generator will evaluated, does
        not have to be separated.
    '''
    if grid is None:
        grid = field.grid
    else:
        field = Field(field, grid)

    axes_reversed = np.array(grid.separated_coords)
    interp = RegularGridInterpolator(axes_reversed, field.shaped, 'nearest', False)

    def interpolator(evaluated_grid):
        evaluated_coords = np.flip(np.array(evaluated_grid.coords), 0)
        res = interp(evaluated_coords.T)
        return Field(res.ravel(), evaluated_grid)

    return interpolator

def make_nearest_interpolator_unstructured(field, grid=None):
    '''Make a nearest interpolator for an unstructured grid.

    Parameters
    ----------
    field : Field or array_like
        The field to interpolate.
    grid : Grid or None
        The grid of the field. If it is given, the grid of `field` is replaced by this grid.

    Returns
    -------
    Field generator
        The interpolator as a Field generator. The grid on which this field generator will be evaluated does
        not need to have any structure.
    '''
    if grid is None:
        grid = field.grid
    else:
        field = Field(field, grid)

    interp = NearestNDInterpolator(grid.points, field)

    def interpolator(evaluated_grid):
        res = interp(grid.points)
        return Field(res, evaluated_grid)

    return interpolator

def make_nearest_interpolator(field, grid=None):
    '''Make a nearest interpolator for any type of grid.

    Parameters
    ----------
    field : Field or array_like
        The field to interpolate.
    grid : Grid or None
        The grid of the field. If it is given, the grid of `field` is replaced by this grid.

    Returns
    -------
    Field generator
        The interpolator as a Field generator. The grid on which this field generator will be evaluated does
        not need to have any structure.
    '''
    if grid is None:
        grid = field.grid

    if grid.is_unstructured:
        return make_nearest_interpolator_unstructured(field, grid)
    else:
        return make_nearest_interpolator_separated(field, grid)



from hcipy import *
import numpy as np
import matplotlib.pyplot as plt
import time

from hcipy.atmosphere import *

pupil_grid = make_pupil_grid(256, 1)
focal_grid = make_focal_grid(pupil_grid, 8, 16)

r0 = 0.3
wavelength = 500e-9
L0 = 10
velocity = 3
height = 0
stencil_length = 2
oversampling = 128

mode_basis = make_zernike_basis(500, 1, pupil_grid, 1)

layers = []
layer = InfiniteAtmosphericLayer(pupil_grid, Cn_squared_from_fried_parameter(r0, wavelength), L0, velocity, 0, stencil_length)
layer2 = ModalAdaptiveOpticsLayer(layer, mode_basis, 1)
layers.append(layer2)
layer = InfiniteAtmosphericLayer(pupil_grid, Cn_squared_from_fried_parameter(r0, wavelength), L0, velocity, 0, stencil_length)
layer3 = ModalAdaptiveOpticsLayer(layer, mode_basis, 1)
layers.append(layer3)
layer = InfiniteAtmosphericLayer(pupil_grid, Cn_squared_from_fried_parameter(r0, wavelength), L0, velocity, 0, stencil_length)
layer4 = ModalAdaptiveOpticsLayer(layer, mode_basis, 1)
layers.append(layer4)
layer = InfiniteAtmosphericLayer(pupil_grid, Cn_squared_from_fried_parameter(r0, wavelength), L0, velocity, 0, stencil_length)
layer5 = ModalAdaptiveOpticsLayer(layer, mode_basis, 1)
layers.append(layer5)
layer = InfiniteAtmosphericLayer(pupil_grid, Cn_squared_from_fried_parameter(r0, wavelength), L0, velocity, 0, stencil_length)
layer6 = ModalAdaptiveOpticsLayer(layer, mode_basis, 1)
layers.append(layer6)

atmosphere = MultiLayerAtmosphere(layers, False)
atmosphere.Cn_squared = Cn_squared_from_fried_parameter(1/40, wavelength)
prop = FraunhoferPropagator(pupil_grid, focal_grid.scaled(wavelength))

aperture = make_circular_aperture(1)(pupil_grid)
wf = Wavefront(Field(np.ones(pupil_grid.size), pupil_grid), wavelength)

for t in np.linspace(0, 100, 5001):
    atmosphere.evolve_until(t)

    wf2 = atmosphere.forward(wf)
    wf2.electric_field *= aperture
    img = Field(prop(wf2).intensity, focal_grid)

    plt.clf()
    plt.subplot(1,2,1)
    imshow_field(wf2.phase, cmap='RdBu')
    plt.subplot(1,2,2)
    imshow_field(np.log10(img / img.max()), vmin=-6)
    plt.draw()
    plt.pause(0.00001)



from __future__ import division

from .atmospheric_model import AtmosphericLayer, phase_covariance_von_karman, fried_parameter_from_Cn_squared
from ..field import Field, RegularCoords, UnstructuredCoords, CartesianGrid
from .finite_atmospheric_layer import FiniteAtmosphericLayer

import numpy as np
from scipy import linalg
from scipy.ndimage import affine_transform

import warnings
import copy

class InfiniteAtmosphericLayer(AtmosphericLayer):
    '''An atmospheric layer that can be infinitely extended in any direction.

    This is an implementation of [Assemat2006]_. Contrary to the
    FiniteAtmosphericLayer class, this atmospheric layer uses an
    autoregressive model for extending the phase screen by one
    column or row in any direction. Each new row/column is based on
    a  stencil, which contains the last few rows/columns in addition
    to a few pixels spread out more into the phase screen to capture
    the low-order aberrations.

    .. [Assemat2006] François Assémat, Richard W. Wilson, and Eric
        Gendron, "Method for simulating infinitely long and non
        stationary phase screens with optimized memory storage,"
        Opt. Express 14, 988-999 (2006)

    .. note::
        This algorithm does not work well with large outer scales.
        This is an inherent flaw in the algorithm and so cannot be avoided
        easily. It requires extremely large-scale correlations to be included
        in the correlation matrices. These correlation matrices are inverted
        by the algorithm, leading to ill-defined inversions.

    Parameters
    ----------
    input_grid : Grid
        The grid on which the incoming wavefront is defined.
    Cn_squared : scalar
        The integrated strength of the turbulence for this layer.
    L0 : scalar
        The outer scale for the atmospheric turbulence of this layer.
        The default is infinity.
    velocity : scalar or array_like
        The wind speed for this atmospheric layer. If this is a scalar,
        the wind will be along x. If this is a 2D array, then the values
        are interpreted as the wind speed along x and y. The default is
        zero.
    height : scalar
        The height of the atmospheric layer. By itself, this value has no
        influence, but it'll be used by the AtmosphericModel to perform
        inter-layer propagations.
    stencil_length : integer
        The number of columns/rows to use for the extrapolation of the phase
        along a column/row. Additionally, the stencil will contain one more
        pixel a little away from these initial columns/rows. Note, do not
        modify this parameter unless you know what you are doing. This parameter
        can be unintuitive. The default value of 2 is usually sufficient in
        most instances.
    use_interpolation : boolean
        whether to use sub-pixel interpolation of the phase screen. Bilinear
        interpolation is used. Without interpolation, for short integration times
        or fast loop speeds, the phase screen may stay on the same pixel for
        multiple frames. With interpolation, these discrete pixel transitions
        are smoothed out. The default is True.
    seed : None, int, array of ints, SeedSequence, BitGenerator, Generator
        A seed to initialize the spectral noise. If None, then fresh, unpredictable
        entry will be pulled from the OS. If an int or array of ints, then it will
        be passed to a numpy.SeedSequency to derive the initial BitGenerator state.
        If a BitGenerator or Generator are passed, these will be wrapped and used
        instead. Default: None.

    Raises
    ------
    ValueError
        When the input grid is not cartesian, regularly spaced and two-dimensional.
    '''
    def __init__(self, input_grid, Cn_squared, L0=np.inf, velocity=0, height=0, stencil_length=2, use_interpolation=True, seed=None):
        self._initialized = False

        AtmosphericLayer.__init__(self, input_grid, Cn_squared, L0, velocity, height)

        # Check properties of input_grid
        if not input_grid.is_('cartesian'):
            raise ValueError('Input grid must be cartesian.')
        if not input_grid.is_regular:
            raise ValueError('Input grid must be regularly spaced')
        if not input_grid.ndim == 2:
            raise ValueError('Input grid must be two-dimensional.')

        self.stencil_length = stencil_length
        self.use_interpolation = use_interpolation
        self.rng = np.random.default_rng(seed)

        self._make_stencils()
        self._make_covariance_matrices()
        self._make_ab_matrices()

        self._original_rng = self.rng
        self.reset()

        self._initialized = True

    def _recalculate_matrices(self):
        if self._initialized:
            self._make_covariance_matrices()
            self._make_ab_matrices()

    def _make_stencils(self):
        # Vertical
        zero = self.input_grid.zero - np.array([0, self.input_grid.delta[1]])
        self.new_grid_bottom = CartesianGrid(RegularCoords(self.input_grid.delta, [self.input_grid.dims[0], 1], zero))

        self.stencil_bottom = Field(np.zeros(self.input_grid.size, dtype='bool'), self.input_grid).shaped
        self.stencil_bottom[:self.stencil_length, :] = True

        for i, n in enumerate(self.rng.geometric(0.5, self.input_grid.dims[0])):
            self.stencil_bottom[(n + self.stencil_length - 1) % self.input_grid.dims[1], i] = True

        self.stencil_bottom = self.stencil_bottom.ravel()
        self.num_stencils_vertical = np.sum(self.stencil_bottom)

        # Horizontal
        zero = self.input_grid.zero - np.array([self.input_grid.delta[0], 0])
        self.new_grid_left = CartesianGrid(RegularCoords(self.input_grid.delta, [1, self.input_grid.dims[1]], zero))

        self.stencil_left = Field(np.zeros(self.input_grid.size, dtype='bool'), self.input_grid).shaped
        self.stencil_left[:, :self.stencil_length] = True

        for i, n in enumerate(self.rng.geometric(0.5, self.input_grid.dims[1])):
            self.stencil_left[i, (n + self.stencil_length - 1) % self.input_grid.dims[0]] = True

        self.stencil_left = self.stencil_left.ravel()
        self.num_stencils_horizontal = np.sum(self.stencil_left)

    def _make_covariance_matrices(self):
        phase_covariance = phase_covariance_von_karman(fried_parameter_from_Cn_squared(1, 1), self.L0)

        # Vertical
        x = np.concatenate((self.input_grid.x[self.stencil_bottom], self.new_grid_bottom.x))
        x = np.concatenate([x - xx for xx in x])
        y = np.concatenate((self.input_grid.y[self.stencil_bottom], self.new_grid_bottom.y))
        y = np.concatenate([y - yy for yy in y])

        separations = CartesianGrid(UnstructuredCoords((x, y)))
        n = self.new_grid_bottom.size + self.num_stencils_vertical
        self.cov_matrix_vertical = phase_covariance(separations).reshape((n, n))

        # Horizontal
        x = np.concatenate((self.input_grid.x[self.stencil_left], self.new_grid_left.x))
        x = np.concatenate([x - xx for xx in x])
        y = np.concatenate((self.input_grid.y[self.stencil_left], self.new_grid_left.y))
        y = np.concatenate([y - yy for yy in y])

        separations = CartesianGrid(UnstructuredCoords((x, y)))
        n = self.new_grid_left.size + self.num_stencils_horizontal
        self.cov_matrix_horizontal = phase_covariance(separations).reshape((n, n))

    def _make_ab_matrices(self):
        # Vertical
        n = self.num_stencils_vertical
        cov_zz = self.cov_matrix_vertical[:n, :n]
        cov_xz = self.cov_matrix_vertical[n:, :n]
        cov_zx = self.cov_matrix_vertical[:n, n:]
        cov_xx = self.cov_matrix_vertical[n:, n:]

        cf = linalg.cho_factor(cov_zz)
        inv_cov_zz = linalg.cho_solve(cf, np.eye(cov_zz.shape[0]))

        self.A_vertical = cov_xz.dot(inv_cov_zz)

        BBt = cov_xx - self.A_vertical.dot(cov_zx)

        U, S, Vt = np.linalg.svd(BBt)
        L = np.sqrt(S[:self.input_grid.dims[0]])

        self.B_vertical = U * L

        # Horizontal
        n = self.num_stencils_horizontal
        cov_zz = self.cov_matrix_horizontal[:n, :n]
        cov_xz = self.cov_matrix_horizontal[n:, :n]
        cov_zx = self.cov_matrix_horizontal[:n, n:]
        cov_xx = self.cov_matrix_horizontal[n:, n:]

        cf = linalg.cho_factor(cov_zz)
        inv_cov_zz = linalg.cho_solve(cf, np.eye(cov_zz.shape[0]))

        self.A_horizontal = cov_xz.dot(inv_cov_zz)

        BBt = cov_xx - self.A_horizontal.dot(cov_zx)

        U, S, Vt = np.linalg.svd(BBt)
        L = np.sqrt(S[:self.input_grid.dims[1]])

        self.B_horizontal = U * L

    def _make_initial_phase_screen(self):
        oversampling = 16

        layer = FiniteAtmosphericLayer(self.input_grid, self.Cn_squared, self.outer_scale, self.velocity, self.height, oversampling, self.rng)

        self._achromatic_screen = layer.phase_for(1)
        self._shifted_achromatic_screen = self._achromatic_screen

        # Reuse the rng from the temporary layer, since it "forwarded" the randomness.
        # This avoids reusing the same randomness every call to reset().
        self.rng = layer.rng

    def _extrude(self, where=None):
        flipped = (where == 'top') or (where == 'right')
        horizontal = (where == 'left') or (where == 'right')

        if where == 'top' or where == 'right':
            screen = self._achromatic_screen[::-1]
        else:
            screen = self._achromatic_screen

        if horizontal:
            stencil = self.stencil_left
            A = self.A_horizontal
            B = self.B_horizontal
        else:
            stencil = self.stencil_bottom
            A = self.A_vertical
            B = self.B_vertical

        stencil_data = screen[stencil]
        random_data = self.rng.normal(0, 1, size=B.shape[1])
        new_slice = A.dot(stencil_data) + B.dot(random_data) * np.sqrt(self._Cn_squared)

        screen = screen.shaped

        if horizontal:
            screen = np.hstack((new_slice[:, np.newaxis], screen[:, :-1]))
        else:
            screen = np.vstack((new_slice[np.newaxis, :], screen[:-1, :]))

        screen = Field(screen, self.input_grid)

        if flipped:
            self._achromatic_screen = screen[::-1, ::-1].ravel()
        else:
            self._achromatic_screen = screen.ravel()

    def phase_for(self, wavelength):
        '''Compute the phase at a certain wavelength.

        Parameters
        ----------
        wavelength : scalar
            The wavelength of the light for which to compute the phase screen.

        Returns
        -------
        Field
            The computed phase screen.
        '''
        return self._shifted_achromatic_screen / wavelength

    def reset(self, make_independent_realization=False):
        '''Reset the atmospheric layer to t=0.

        Parameters
        ----------
        make_independent_realization : boolean
            Whether to start an independent realization of the noise for the
            atmospheric layer or not. When this is False, the exact same phase
            screens will be generated as in the first run, as long as the Cn^2
            and outer scale are the same. This allows for testing of multiple
            runs of a control system with different control parameters. When
            this is True, an independent realization of the atmospheric layer
            will be generated. This is useful for Monte-Carlo-style computations.
            The default is False.
        '''
        if make_independent_realization:
            # Reset the original random generator to the current one. This
            # will essentially reset the randomness.
            self._original_rng = copy.copy(self.rng)
        else:
            # Make a copy of the original random generator. This copy will be
            # used as the source for all randomness.
            self.rng = copy.copy(self._original_rng)

        self._make_initial_phase_screen()

        self.center = np.zeros(2)
        self._t = 0

    def evolve_until(self, t):
        '''Evolve the atmospheric layer until a certain time.

        .. note::
            Backwards evolution is not allowed. The information about previous
            parts of the phase screen are inherently lost. If you want to replay the
            same atmospheric noise again, please use `reset()` to reset the layer
            to its starting position.

        Parameters
        ----------
        t : scalar
            The new time to evolve the phase screen to. This should be larger or equal to
            the current time of the atmospheric layer.

        Raises
        ------
        ValueError
            When the time `t` is smaller than the current time of the layer.
        '''
        if t is None:
            self.reset()
            return

        if t < self._t:
            raise ValueError('Backwards temporal evolution is not allowed.')

        # Get the old and new center positions
        old_center = np.round(self.center / self.input_grid.delta).astype('int')
        self.center = self.center + self.velocity * (t - self._t)
        new_pixel_center = np.round(self.center / self.input_grid.delta).astype('int')

        self._t = t

        # Measure the number of full pixel shifts
        delta = new_pixel_center - old_center
        # Measure the sub-pixel shift
        sub_delta = self.center - new_pixel_center * self.input_grid.delta

        for i in range(abs(delta[0])):
            if delta[0] < 0:
                self._extrude('left')
            else:
                self._extrude('right')

        for i in range(abs(delta[1])):
            if delta[1] < 0:
                self._extrude('bottom')
            else:
                self._extrude('top')

        if self.use_interpolation:
            # Use bilinear interpolation to interpolate the achromatic phase screen to the correct position.
            # This is to avoid sudden shifts by discrete pixels.
            ps = self._achromatic_screen.shaped

            with warnings.catch_warnings():
                # Suppress warnings about the changed behaviour in affine_transform for 1D arrays.
                # We know about this and are expecting the behaviour as is.
                warnings.filterwarnings('ignore', message='The behaviour of affine_transform')
                warnings.filterwarnings('ignore', message='The behavior of affine_transform')

                screen = affine_transform(ps, np.array([1, 1]), (sub_delta / self.input_grid.delta)[::-1], mode='nearest', order=5)
                self._shifted_achromatic_screen = Field(screen.ravel(), self._achromatic_screen.grid)
        else:
            self._shifted_achromatic_screen = self._achromatic_screen

    @property
    def Cn_squared(self):  # noqa: N802
        '''The integrated strength of the turbulence for this layer.
        '''
        return self._Cn_squared

    @Cn_squared.setter
    def Cn_squared(self, Cn_squared):  # noqa: N802
        self._Cn_squared = Cn_squared

    @property
    def outer_scale(self):
        '''The outer scale of the turbulence for this layer.
        '''
        return self._L0

    @outer_scale.setter
    def L0(self, L0):  # noqa: N802
        self._L0 = L0

        self._recalculate_matrices()



import numpy as np
import copy

class Coords(object):
    '''Base class for coordinates.
    '''
    _coordinate_types = {}

    def copy(self):
        '''Make a copy.
        '''
        return copy.deepcopy(self)

    @classmethod
    def from_dict(cls, tree):
        coordinate_class = Coords._coordinate_types[tree['type']]

        return coordinate_class.from_dict(tree)

    def to_dict(self):
        raise NotImplementedError()

    def __add__(self, b):
        '''Add `b` to the coordinates separately and return the result.
        '''
        res = self.copy()
        res += b
        return res

    def __iadd__(self, b):
        '''Add `b` to the coordinates separately in-place.
        '''
        raise NotImplementedError()

    def __radd__(self, b):
        '''Add `b` to the coordinates separately and return the result.
        '''
        return self + b

    def __sub__(self, b):
        '''Subtract `b` from the coordinates separately and return the result.
        '''
        return self + (-b)

    def __isub__(self, b):
        '''Subtract `b` from the coordinates separately in-place.
        '''
        self += (-b)
        return self

    def __mul__(self, f):
        '''Multiply each coordinate with `f` separately and return the result.
        '''
        res = self.copy()
        res *= f
        return res

    def __rmul__(self, f):
        '''Multiply each coordinate with `f` separately and return the result.
        '''
        return self * f

    def __imul__(self, f):
        '''Multiply each coordinate with `f` separately in-place.
        '''
        raise NotImplementedError()

    def __div__(self, f):
        '''Divide each coordinate with `f` separately and return the result.
        '''
        return self * (1 / f)

    def __idiv__(self, f):
        '''Divide each coordinate with `f` separately in-place.
        '''
        self *= 1 / f
        return self

    def __eq__(self, other):
        '''Check if the coordinates are identical.

        Parameters
        ----------
        other : object
            The object to which to compare.

        Returns
        -------
        boolean
            Whether the two objects are identical.
        '''
        return NotImplemented

    def __getitem__(self, i):
        '''The `i`-th point for these coordinates.
        '''
        raise NotImplementedError()

    @property
    def is_separated(self):
        '''True if the coordinates are separated, False otherwise.
        '''
        return hasattr(self, 'separated_coords')

    @property
    def is_regular(self):
        '''True if the coordinates are regularly-spaced, False otherwise.
        '''
        return hasattr(self, 'regular_coords')

    @property
    def is_unstructured(self):
        '''True if the coordinates are not structured, False otherwise.
        '''
        return not self.is_separated

    def reverse(self):
        '''Reverse the ordering of points in-place.
        '''
        raise NotImplementedError()

    @property
    def size(self):
        '''The number of points.
        '''
        raise NotImplementedError()

    def __len__(self):
        '''The number of dimensions.
        '''
        raise NotImplementedError()

    @classmethod
    def _add_coordinate_type(cls, coordinate_type, coordinate_class):
        cls._coordinate_types[coordinate_type] = coordinate_class

class UnstructuredCoords(Coords):
    '''An unstructured list of points.

    Parameters
    ----------
    coords : list or tuple
        A tuple of a list of positions for each dimension.
    '''
    def __init__(self, coords):
        self.coords = [np.array(c) for c in coords]

    @classmethod
    def from_dict(cls, tree):
        '''Make an UnstructuredCoords from a dictionary, previously created by `to_dict()`.

        Parameters
        ----------
        tree : dictionary
            The dictionary from which to make a new UnstructuredCoords object.

        Returns
        -------
        UnstructuredCoords
            The created object.

        Raises
        ------
        ValueError
            If the dictionary is not formatted correctly.
        '''
        if tree['type'] != 'unstructured':
            raise ValueError('The type of coordinates should be "unstructured".')

        return cls(tree['coords'])

    def to_dict(self):
        '''Convert the object to a dictionary for serialization.

        Returns
        -------
        dictionary
            The created dictionary.
        '''
        tree = {
            'type': 'unstructured',
            'coords': self.coords
        }

        return tree

    @property
    def size(self):
        '''The number of points.
        '''
        return self.coords[0].size

    def __len__(self):
        '''The number of dimensions.
        '''
        return len(self.coords)

    def __getitem__(self, i):
        '''The `i`-th point for these coordinates.
        '''
        return self.coords[i]

    def __iadd__(self, b):
        '''Add `b` to the coordinates separately in-place.
        '''
        b = np.ones(len(self.coords)) * b
        for i in range(len(self.coords)):
            self.coords[i] += b[i]
        return self

    def __imul__(self, f):
        '''Multiply each coordinate with `f` separately in-place.
        '''
        f = np.ones(len(self.coords)) * f
        for i in range(len(self.coords)):
            self.coords[i] *= f[i]
        return self

    def __eq__(self, other):
        '''Check if the coordinates are identical.

        Parameters
        ----------
        other : object
            The object to which to compare.

        Returns
        -------
        boolean
            Whether the two objects are identical.
        '''
        if type(self) != type(other):
            return False

        return np.array_equal(self.coords, other.coords)

    def reverse(self):
        '''Reverse the ordering of points in-place.
        '''
        for i in range(len(self.coords)):
            self.coords[i] = self.coords[i][::-1]
        return self

class SeparatedCoords(Coords):
    '''A list of points that are separable along each dimension.

    The actual points are given by the iterated tensor product of the `separated_coords`.

    Parameters
    ----------
    separated_coords : list or tuple
        A tuple of a list of coordinates along each dimension.

    Attributes
    ----------
    separated_coords
        A tuple of a list of coordinates along each dimension.
    '''
    def __init__(self, separated_coords):
        # Make a copy to avoid modification from outside the class
        self.separated_coords = [copy.deepcopy(s) for s in separated_coords]

    @classmethod
    def from_dict(cls, tree):
        '''Make an SeparatedCoords from a dictionary, previously created by `to_dict()`.

        Parameters
        ----------
        tree : dictionary
            The dictionary from which to make a new SeparatedCoords object.

        Returns
        -------
        SeparatedCoords
            The created object.

        Raises
        ------
        ValueError
            If the dictionary is not formatted correctly.
        '''
        if tree['type'] != 'separated':
            raise ValueError('The type of coordinates should be "separated".')

        return cls(tree['separated_coords'])

    def to_dict(self):
        '''Convert the object to a dictionary for serialization.

        Returns
        -------
        dictionary
            The created dictionary.
        '''
        tree = {
            'type': 'separated',
            'separated_coords': self.separated_coords
        }

        return tree

    def __getitem__(self, i):
        '''The `i`-th point for these coordinates.
        '''
        s0 = (1,) * len(self)
        j = len(self) - i - 1
        output = self.separated_coords[i].reshape(s0[:j] + (-1,) + s0[j + 1:])
        return np.broadcast_to(output, self.shape).ravel()

    @property
    def size(self):
        '''The number of points.
        '''
        return np.prod(self.shape)

    def __len__(self):
        '''The number of dimensions.
        '''
        return len(self.separated_coords)

    @property
    def dims(self):
        '''The number of points along each dimension.
        '''
        return np.array([len(c) for c in self.separated_coords])

    @property
    def shape(self):
        '''The shape of an ``numpy.ndarray`` with the right dimensions.
        '''
        return self.dims[::-1]

    def __iadd__(self, b):
        '''Add `b` to the coordinates separately in-place.
        '''
        for i in range(len(self)):
            self.separated_coords[i] += b[i]
        return self

    def __imul__(self, f):
        '''Multiply each coordinate with `f` separately in-place.
        '''
        if np.isscalar(f):
            for i in range(len(self)):
                self.separated_coords[i] *= f
        else:
            for i in range(len(self)):
                self.separated_coords[i] *= f[i]
        return self

    def __eq__(self, other):
        '''Check if the coordinates are identical.

        Parameters
        ----------
        other : object
            The object to which to compare.

        Returns
        -------
        boolean
            Whether the two objects are identical.
        '''
        if type(self) != type(other):
            return False

        return np.array_equal(self.separated_coords, other.separated_coords)

    def reverse(self):
        '''Reverse the ordering of points in-place.
        '''
        for i in range(len(self)):
            self.separated_coords[i] = self.separated_coords[i][::-1]
        return self

class RegularCoords(Coords):
    '''A list of points that have a regular spacing in all dimensions.

    Parameters
    ----------
    delta : array_like
        The spacing between the points.
    dims : array_like
        The number of points along each dimension.
    zero : array_like
        The coordinates for the first point.

    Attributes
    ----------
    delta
        The spacing between the points.
    dims
        The number of points along each dimension.
    zero
        The coordinates for the first point.
    '''
    def __init__(self, delta, dims, zero=None):
        if np.isscalar(dims):
            self.dims = np.array([dims]).astype('int')
        else:
            self.dims = np.array(dims).astype('int')

        if np.isscalar(delta):
            self.delta = np.array([delta] * len(self.dims))
        else:
            self.delta = np.array(delta)

        if zero is None:
            self.zero = np.zeros(len(self.dims))
        elif np.isscalar(zero):
            self.zero = np.array([zero] * len(self.dims))
        else:
            self.zero = np.array(zero)

    @classmethod
    def from_dict(cls, tree):
        '''Make an RegularCoords from a dictionary, previously created by `to_dict()`.

        Parameters
        ----------
        tree : dictionary
            The dictionary from which to make a new RegularCoords object.

        Returns
        -------
        RegularCoords
            The created object.

        Raises
        ------
        ValueError
            If the dictionary is not formatted correctly.
        '''
        if tree['type'] != 'regular':
            raise ValueError('The type of coordinates should be "regular".')

        return cls(tree['delta'], tree['dims'], tree['zero'])

    def to_dict(self):
        '''Convert the object to a dictionary for serialization.

        Returns
        -------
        dictionary
            The created dictionary.
        '''
        tree = {
            'type': 'regular',
            'delta': self.delta.tolist(),
            'dims': self.dims.tolist(),
            'zero': self.zero.tolist()
        }

        return tree

    @property
    def separated_coords(self):
        '''A tuple of a list of the values for each dimension.

        The actual points are the iterated tensor product of this tuple.
        '''
        return [np.arange(n) * delta + zero for delta, n, zero in zip(self.delta, self.dims, self.zero)]

    @property
    def regular_coords(self):
        '''The tuple `(delta, dims, zero)` of the regularly-spaced coordinates.
        '''
        return self.delta, self.dims, self.zero

    @property
    def size(self):
        '''The number of points.
        '''
        return np.prod(self.dims)

    def __len__(self):
        '''The number of dimensions.
        '''
        return len(self.dims)

    @property
    def shape(self):
        '''The shape of an ``numpy.ndarray`` with the right dimensions.
        '''
        return self.dims[::-1]

    def __getitem__(self, i):
        '''The `i`-th point for these coordinates.
        '''
        s0 = (1,) * len(self)
        j = len(self) - i - 1
        t = s0[:j] + (-1,) + s0[j + 1:]
        output = self.separated_coords[i].reshape(t)
        return np.broadcast_to(output, self.shape).ravel()

    def __iadd__(self, b):
        '''Add `b` to the coordinates separately in-place.
        '''
        self.zero += b
        return self

    def __imul__(self, f):
        '''Multiply each coordinate with `f` separately in-place.
        '''
        self.delta *= f
        self.zero *= f
        return self

    def __eq__(self, other):
        '''Check if the coordinates are identical.

        Parameters
        ----------
        other : object
            The object to which to compare.

        Returns
        -------
        boolean
            Whether the two objects are identical.
        '''
        if type(self) != type(other):
            return False

        return np.array_equal(self.regular_coords, other.regular_coords)

    def reverse(self):
        '''Reverse the ordering of points in-place.
        '''
        maximum = self.zero + self.delta * (self.dims - 1)
        self.delta = -self.delta
        self.zero = maximum
        return self

Coords._add_coordinate_type('unstructured', UnstructuredCoords)
Coords._add_coordinate_type('separated', SeparatedCoords)
Coords._add_coordinate_type('regular', RegularCoords)



import numpy as np
from ..field import Field

class FourierTransform(object):
    '''The base class for all Fourier transform implementations.

    Attributes
    ----------
    input_grid : Grid
        The grid that is expected for the input field.
    output_grid : Grid
        The grid that is produced by the Fourier transform.
    '''
    def forward(self, field):
        '''Returns the forward Fourier transform of the :class:`Field` field.

        Parameters
        ----------
        field : Field
            The field to Fourier transform.

        Returns
        -------
        Field
            The Fourier transform of the field.
        '''
        raise NotImplementedError()

    def backward(self, field):
        '''Returns the inverse Fourier transform of the :class:`Field` field.

        Parameters
        ----------
        field : Field
            The field to inverse Fourier transform.

        Returns
        -------
        Field
            The inverse Fourier transform of the field.
        '''
        raise NotImplementedError()

    def get_transformation_matrix_forward(self):
        '''Returns the transformation matrix corresonding to the
        Fourier transform.

        Returns
        -------
        ndarray
            A matrix representing the Fourier transform.
        '''
        coords_in = self.input_grid.as_('cartesian').coords
        coords_out = self.output_grid.as_('cartesian').coords

        A = np.exp(-1j * np.dot(np.array(coords_out).T, coords_in))
        A *= self.input_grid.weights

        return A

    def get_transformation_matrix_backward(self):
        '''Returns the transformation matrix corresonding to the
        Fourier transform.

        Returns
        -------
        ndarray
            A matrix representing the Fourier transform.
        '''
        coords_in = self.input_grid.as_('cartesian').coords
        coords_out = self.output_grid.as_('cartesian').coords

        A = np.exp(1j * np.dot(np.array(coords_in).T, coords_out))
        A *= self.output_grid.weights
        A /= (2 * np.pi)**self.input_grid.ndim

        return A

def _time_it_iterative(function, num_iterations):
    import time

    start = time.perf_counter()
    for _ in range(num_iterations):
        function()
    end = time.perf_counter()

    return (end - start) / num_iterations

def _time_it(function, t_max=0.1, repeat_max=5):
    num_iterations = 1

    while True:
        time_per_iteration = _time_it_iterative(function, num_iterations)

        if time_per_iteration * num_iterations > t_max:
            break
        else:
            num_iterations *= 2

    # Shortcut if one iteration of the function itself takes longer than repeat_max * t_max.
    if num_iterations == 1 and time_per_iteration > (repeat_max * t_max):
        return time_per_iteration

    times = [time_per_iteration]

    for _ in range(repeat_max - 1):
        times.append(_time_it_iterative(function, num_iterations))

    return np.median(times)

def make_fourier_transform(input_grid, output_grid=None, q=1, fov=1, shift=0, planner='estimate'):
    '''Construct a FourierTransform object.

    The most time-efficient Fourier transform method will be chosen according to actual or estimated performance.

    Parameters
    ----------
    input_grid : Grid
        The grid that will be used for the Field passed to the Fourier transform.
    output_grid : None or Grid
        The grid of the resulting field. If it is None, a optimal grid will be chosen, according to `q` and `fov`.
    q : scalar or ndarray
        Describes how many samples to take in the Fourier domain. A value of 1 means critcally sampled in the Fourier domain.
    fov : scalar or ndarray
        Describes how far out the Fourier domain extends. A value of 1 means the same amount of samples as the spatial domain.
    shift : scalar or ndarray
        Describes by how much the Fourier domain should be shifted compared to the native sampling of FFT.
    planner : string
        If it is 'estimate', performance of the different methods will be estimated from theoretical complexity estimates.
        If it is 'measure', actual Fourier transforms will be performed to get the actual performance. The latter takes longer,
        but is more accurate.

    Returns
    -------
    FourierTransform
        The Fourier transform that was requested.
    '''
    from .fast_fourier_transform import FastFourierTransform, make_fft_grid, get_fft_parameters
    from .matrix_fourier_transform import MatrixFourierTransform
    from .naive_fourier_transform import NaiveFourierTransform

    if output_grid is not None:
        # Try to detect if the grid is compatible with an FFT grid.
        try:
            q, fov, shift = get_fft_parameters(output_grid, input_grid)

            # If we got this far, the output grid is a native FFT grid.
            # Remove output grid as its now completely defined by q, fov and shift parameters.
            output_grid = None
        except ValueError:
            # The grid is not a native FFT grid.
            pass

    if output_grid is None:
        # Choose between FFT and MFT
        if not (input_grid.is_regular and input_grid.is_('cartesian')):
            raise ValueError('For non-regular non-cartesian Grids, a Fourier transform is required to have an output_grid.')

        if input_grid.ndim not in [1, 2]:
            method = 'fft'
        else:
            output_grid = make_fft_grid(input_grid, q, fov, shift)

            if planner == 'estimate':
                # Estimate analytically from complexities.
                # Convert shapes to float to avoid potential overflows.
                N_in = input_grid.shape.astype('float') * q
                N_out = output_grid.shape.astype('float')

                if input_grid.ndim == 1:
                    fft = 4 * N_in[0] * np.log2(N_in)
                    mft = 4 * input_grid.size * N_out[0]
                else:
                    fft = 4 * np.prod(N_in) * np.log2(np.prod(N_in))
                    mft = 4 * (np.prod(input_grid.shape) * N_out[1] + np.prod(N_out) * input_grid.shape[0])

                if fft > mft:
                    method = 'mft'
                else:
                    method = 'fft'
            elif planner == 'measure':
                # Measure directly.
                fft = FastFourierTransform(input_grid, q, fov)
                mft = MatrixFourierTransform(input_grid, output_grid)

                a = input_grid.zeros(dtype='complex')
                fft_time = _time_it(lambda: fft.forward(a))
                mft_time = _time_it(lambda: mft.forward(a))

                if fft_time > mft_time:
                    method = 'mft'
                else:
                    method = 'fft'
    else:
        # Choose between MFT and Naive
        if input_grid.is_separated and input_grid.is_('cartesian') and output_grid.is_separated and output_grid.is_('cartesian') and input_grid.ndim in [1, 2]:
            method = 'mft'
        else:
            method = 'naive'

    # Make the Fourier transform
    if method == 'fft':
        return FastFourierTransform(input_grid, q, fov, shift)
    elif method == 'mft':
        return MatrixFourierTransform(input_grid, output_grid)
    elif method == 'naive':
        return NaiveFourierTransform(input_grid, output_grid)

def multiplex_for_tensor_fields(func):
    '''A decorator for automatically multiplexing a function over the tensor directions.

    This function is used internally for simplifying the implementation of the Fourier transforms.

    Parameters
    ----------
    func : function
        The function to multiplex. This function gets called for each of the tensor elements.
    '''
    def inner(self, field):
        if field.is_scalar_field:
            return func(self, field)
        else:
            f = field.reshape((-1, field.grid.size))
            res = [func(self, ff) for ff in f]
            new_shape = np.concatenate((field.tensor_shape, [-1]))
            return Field(np.array(res).reshape(new_shape), res[0].grid)

    return inner

def _get_float_and_complex_dtype(dtype):
    '''Return the floating point and complex numpy data types with the same
    bit depth per component.

    Parameters
    ----------
    dtype : numpy dtype
        The dtype to convert to a floating point and complex dtype.

    Returns
    -------
    float_dtype : numpy dtype
        The floating point dtype.
    complex_dtype : numpy dtype
        The complex dtype.
    '''
    if dtype == np.dtype('float32') or dtype == np.dtype('complex64'):
        complex_dtype = 'complex64'
        float_dtype = 'float32'
    else:
        complex_dtype = 'complex128'
        float_dtype = 'float64'

    return float_dtype, complex_dtype



import pytest

def pytest_addoption(parser):
    parser.addoption('--runslow', action='store_true', help='run slow tests')

def pytest_runtest_setup(item):
    if 'slow' in item.keywords and not item.config.getoption('--runslow'):
        pytest.skip('Need --runslow option to run.')

def pytest_configure(config):
    config.addinivalue_line('markers', 'slow: marks tests as slow')



__all__ = [
    'circular_aperture',
    'make_circular_aperture',
    'elliptical_aperture',
    'make_elliptical_aperture',
    'rectangular_aperture',
    'make_rectangular_aperture',
    'irregular_polygon_aperture',
    'make_irregular_polygon_aperture',
    'regular_polygon_aperture',
    'make_regular_polygon_aperture',
    'hexagonal_aperture',
    'make_hexagonal_aperture',
    'make_spider',
    'make_spider_infinite',
    'make_obstructed_circular_aperture',
    'make_obstruction',
    'make_rotated_aperture',
    'make_shifted_aperture',
    'make_segmented_aperture',
    'make_elt_aperture',
    'make_gmt_aperture',
    'make_tmt_aperture',
    'make_hexagonal_segmented_aperture',
    'make_magellan_aperture',
    'make_luvoir_a_aperture',
    'make_luvoir_a_lyot_stop',
    'make_luvoir_b_aperture',
    'make_hale_aperture',
    'make_hicat_aperture',
    'make_hicat_lyot_stop',
    'make_vlt_aperture',
    'make_vlti_aperture',
    'make_vlti_dopd_map',
    'make_habex_aperture',
    'make_hst_aperture',
    'make_jwst_aperture',
    'make_keck_aperture',
]

from .generic import *
from .realistic import *



__all__ = [
    'make_power_law_error',
    'SurfaceAberration',
    'SurfaceAberrationAtDistance',
    'Apodizer',
    'PhaseApodizer',
    'SurfaceApodizer',
    'ComplexSurfaceApodizer',
    'MultiplexedComplexSurfaceApodizer',
    'DynamicOpticalSystem',
    'make_actuator_positions',
    'make_gaussian_influence_functions',
    'make_xinetics_influence_functions',
    'find_illuminated_actuators',
    'DeformableMirror',
    'label_actuator_centroid_positions',
    'Detector',
    'NoiselessDetector',
    'NoisyDetector',
    'make_gaussian_fiber_mode',
    'StepIndexFiber',
    'SingleModeFiber',
    'SingleModeFiberInjection',
    'SingleModeFiberArray',
    'GaussianBeam',
    'make_sellmeier_glass',
    'make_cauchy_glass',
    'get_refractive_index',
    'get_glasses_in_catalogue',
    'Magnifier',
    'EvenAsphereMicroLensArray',
    'SphericalMicroLensArray',
    'MicroLensArray',
    'closest_points',
    'OpticalElement',
    'EmptyOpticalElement',
    'AgnosticOpticalElement',
    'make_agnostic_forward',
    'make_agnostic_backward',
    'make_agnostic_optical_element',
    'OpticalSystem',
    'PeriodicOpticalElement',
    'jones_to_mueller',
    'JonesMatrixOpticalElement',
    'PhaseRetarder',
    'LinearRetarder',
    'CircularRetarder',
    'QuarterWavePlate',
    'HalfWavePlate',
    'GeometricPhaseElement',
    'LinearPolarizer',
    'LinearPolarizingBeamSplitter',
    'CircularPolarizingBeamSplitter',
    'SegmentedDeformableMirror',
    'spherical_surface_sag',
    'parabolic_surface_sag',
    'conical_surface_sag',
    'even_aspheric_surface_sag',
    'TipTiltMirror',
    'SimpleVibration',
    'Wavefront',
    'ThinLens',
    'Prism',
    'ThinPrism',
    'TiltElement',
    'PhaseGrating'
]

from .optical_element import *
from .wavefront import *

from .aberration import *
from .apodization import *
from .deformable_mirror import *
from .detector import *
from .dispersion_optics import *
from .dynamic_optical_system import *
from .fiber import *
from .gaussian_beam import *
from .glass import *
from .magnifier import *
from .micro_lens_array import *
from .periodic_optical_element import *
from .polarization import *
from .segmented_mirror import *
from .surface_profiles import *
from .thin_lens import *
from .tip_tilt_mirror import *
from .vibration import *



from .mode_basis import ModeBasis
from ..field import Field
import numpy as np

_epsilon = 1e-12

def make_cosine_basis(grid, fourier_grid, sort_by_energy=True):
    '''Make a cosine basis.

    Repeated frequencies will not be repeated in this mode basis. This means that opposite points in the `fourier_grid` will be silently ignored.

    Parameters
    ----------
    grid : Grid
        The :class:`Grid` on which to calculate the modes.
    fourier_grid : Grid
        The grid defining all frequencies.
    sort_by_energy : bool
        Whether to sort by increasing energy or not.

    Returns
    -------
    ModeBasis
        The mode basis containing all cosine modes.
    '''
    modes = []
    energies = []
    ignore_list = []

    c = np.array(grid.coords)

    for i, p in enumerate(fourier_grid.points):
        if i in ignore_list:
            continue

        mode = Field(np.cos(np.dot(p, c)), grid)

        # Correctly normalize the non-zero energy mode.
        if not np.dot(p, p) < _epsilon**2:
            mode *= np.sqrt(2)

        modes.append(mode)

        j = fourier_grid.closest_to(-p)

        dist = fourier_grid.points[j] + p
        dist2 = np.dot(dist, dist)

        p_length2 = np.dot(p, p)
        energies.append(p_length2)

        if dist2 < (_epsilon * p_length2):
            ignore_list.append(j)

    if sort_by_energy:
        ind = np.argsort(energies)
        modes = [modes[i] for i in ind]

    return ModeBasis(modes, grid)

def make_sine_basis(grid, fourier_grid, sort_by_energy=True):
    '''Make a sine basis.

    Repeated frequencies will not be repeated in this mode basis. This means that opposite points in the `fourier_grid` will be silently ignored.

    Parameters
    ----------
    grid : Grid
        The :class:`Grid` on which to calculate the modes.
    fourier_grid : Grid
        The grid defining all frequencies.
    sort_by_energy : bool
        Whether to sort by increasing energy or not.

    Returns
    -------
    ModeBasis
        The mode basis containing all sine modes.
    '''
    modes = []
    energies = []
    ignore_list = []

    c = np.array(grid.coords)

    for i, p in enumerate(fourier_grid.points):
        if i in ignore_list:
            continue

        mode = Field(np.sin(np.dot(p, c)) * np.sqrt(2), grid)

        # Filter out the zero energy mode.
        if np.dot(p, p) < _epsilon**2:
            continue

        modes.append(mode)

        j = fourier_grid.closest_to(-p)

        dist = fourier_grid.points[j] + p
        dist2 = np.dot(dist, dist)

        p_length2 = np.dot(p, p)
        energies.append(p_length2)

        if dist2 < (_epsilon * p_length2):
            ignore_list.append(j)

    if sort_by_energy:
        ind = np.argsort(energies)
        modes = [modes[i] for i in ind]

    return ModeBasis(modes, grid)

def make_fourier_basis(grid, fourier_grid, sort_by_energy=True):
    '''Make a Fourier basis.

    Fourier modes this function are defined to be real. This means that for each point, both a sine and cosine mode is returned.

    Repeated frequencies will not be repeated in this mode basis. This means that opposite points in the `fourier_grid` will be silently ignored.

    Parameters
    ----------
    grid : Grid
        The :class:`Grid` on which to calculate the modes.
    fourier_grid : Grid
        The grid defining all frequencies.
    sort_by_energy : bool
        Whether to sort by increasing energy or not.

    Returns
    -------
    ModeBasis
        The mode basis containing all Fourier modes.
    '''
    modes_cos = []
    modes_sin = []
    energies = []
    ignore_list = []

    c = np.array(grid.coords)

    for i, p in enumerate(fourier_grid.points):
        if i in ignore_list:
            continue

        mode_cos = Field(np.cos(np.dot(p, c)), grid)
        mode_sin = Field(np.sin(np.dot(p, c)), grid)

        modes_cos.append(mode_cos)
        modes_sin.append(mode_sin)

        j = fourier_grid.closest_to(-p)

        dist = fourier_grid.points[j] + p
        dist2 = np.dot(dist, dist)

        p_length2 = np.dot(p, p)
        energies.append(p_length2)

        if dist2 < (_epsilon * p_length2):
            ignore_list.append(j)

    if sort_by_energy:
        ind = np.argsort(energies)
        modes_sin = [modes_sin[i] for i in ind]
        modes_cos = [modes_cos[i] for i in ind]
        energies = np.array(energies)[ind]

    modes = []
    for i, E in enumerate(energies):
        # Filter out and correctly normalize zero energy vs non-zero energy modes.
        if E > _epsilon:
            modes.append(modes_cos[i] * np.sqrt(2))
            modes.append(modes_sin[i] * np.sqrt(2))
        else:
            modes.append(modes_cos[i])

    return ModeBasis(modes, grid)

def make_complex_fourier_basis(grid, fourier_grid, sort_by_energy=True):
    '''Make a complex Fourier basis.

    Fourier modes this function are defined to be complex. For each point in `fourier_grid` the complex Fourier mode is contained in the output.

    Parameters
    ----------
    grid : Grid
        The :class:`Grid` on which to calculate the modes.
    fourier_grid : Grid
        The grid defining all frequencies.
    sort_by_energy : bool
        Whether to sort by increasing energy or not.

    Returns
    -------
    ModeBasis
        The mode basis containing all Fourier modes.
    '''
    c = np.array(grid.coords)

    modes = [Field(np.exp(1j * np.dot(p, c)), grid) for p in fourier_grid.points]
    energies = [np.dot(p, p) for p in fourier_grid.points]

    if sort_by_energy:
        ind = np.argsort(energies)
        modes = [modes[i] for i in ind]

    return ModeBasis(modes, grid)



from hcipy import *
import numpy as np
import matplotlib.pyplot as plt

wavelength = 1e-6 # m
D_tel = 100 * wavelength
N_over = 4
F_num = 3
distance = D_tel * F_num
k = 2*np.pi / wavelength

N_samp = 1024
pupil_grid = make_pupil_grid(N_samp, D_tel*N_over)
aperture = make_circular_aperture(D_tel)
focal_grid = make_focal_grid(pupil_grid, 1, 500)

fresnel = FresnelPropagator(pupil_grid, distance)
ASP = AngularSpectrumPropagator(pupil_grid, distance)
props = [fresnel, ASP]

aper = aperture(pupil_grid) * np.exp(-1j * k * pupil_grid.as_('polar').r**2 / (2 * distance))
tilts = ((0, 0), (10, 10), (20, 20))
Ntilts = len(tilts)

for j, tilt in enumerate(tilts):

    phase_tilt = 2*np.pi * (pupil_grid.x / D_tel * tilt[0] + pupil_grid.y / D_tel * tilt[1])
    aper *= np.exp(1j * phase_tilt)

    wf = Wavefront(aper, wavelength)
    wf.total_power = 1

    for i, prop in enumerate(props):
        img = prop(wf)

        plt.subplot(Ntilts, 2, 2*j + i + 1)
        imshow_field(np.log10(img.intensity / img.intensity.max()), vmin=-5)
plt.show()



import numpy as np

try:
    from importlib.resources import files
except ImportError:
    from importlib_resources import files

def make_sellmeier_glass(A, K, L):
    r'''The Sellmeier equation for the dispersion of the refractive index of materials.

    The dispersion relation of the Sellmeier equation is [Sellmeier1872]_,

    .. math:: n^2 = 1 + \sum_{i}{\frac{K_i \lambda^2}{\lambda^2-L_i}}

    .. [Sellmeier1872] W. Sellmeier 1872, Ueber die durch die Aetherschwingungen erregten
        Mitschwingungen der Körpertheilchen und deren Rückwirkung auf die ersteren, besonders
        zur Erklärung der Dispersion und ihrer Anomalien (II. Theil),"  Annalen der Physik und
        Chemie. 223 (11), 386–403 (1872)

    Parameters
    ----------
    A : scalar
        The Sellmeier coefficient
    K : array_like
        The Sellmeier coefficients
    L : array_like
        The Sellmeier coefficients in um**2

    Returns
    -------
    function
        The refractive index profile function
    '''
    def refractive_index(wavelength):
        n_squared = A
        for Ki, Li in zip(K, L):
            n_squared += Ki * (wavelength * 1e6)**2 / ((wavelength * 1e6)**2 - Li)
        return np.sqrt(n_squared)

    return refractive_index

def make_cauchy_glass(coefficients):
    r'''The Cauchy equation for the dispersion of the refractive index of materials.

    The dispersion relation of the Cauchy equation is,

    .. math:: n = \sum_{i=0}{\frac{K_i}{\lambda^{2i}}

    Parameters
    ----------
    coefficients : scalar or array_like
        The Cauchy coefficients in um**(2*i).

    Returns
    -------
    function
        The refractive index profile function
    '''
    def refractive_index(wavelength):
        n = 0
        for i, coef in enumerate(coefficients):
            n += coef / (wavelength * 1e6)**(2 * i)
        return n

    return refractive_index

_glass_catalogue = None

def get_refractive_index(glass_name):
    '''Get the refractive index for one of the glasses in the catalogue.

    Parameters
    ----------
    glass_name : str
        The name of the glass.

    Returns
    -------
    callable
        The refractive index as function of wavelength.
    '''
    if _glass_catalogue is None:
        _build_glass_catalogue()

    if glass_name not in _glass_catalogue:
        # Not found. Try to find suggestion.
        for key in _glass_catalogue.keys():
            if key.lower() == glass_name.lower():
                raise ValueError('The requested glass "%s" was not found in the catalogue. Did you mean "%s"?' % (glass_name, key))

        # No suggestion found, raise
        raise ValueError('The requested glass "%s" was not found in the catalogue.' % glass_name)

    return _glass_catalogue[glass_name]

def get_glasses_in_catalogue():
    '''Get the names of all the glasses in the glass catalogue.

    Returns
    -------
    iterable of strings
        The names of each of the glasses in the catalogue in arbitrary order.
    '''
    if _glass_catalogue is None:
        _build_glass_catalogue()

    return _glass_catalogue.keys()

def _build_glass_catalogue():
    global _glass_catalogue

    _glass_catalogue = {
        'IP_DIP': make_cauchy_glass([1.5273, 6.5456E-3, 2.5345E-4]),
        'SILICA': make_sellmeier_glass(1, [0.6961663, 0.4079426, 0.8974794], [0.0684043**2, 0.1162414**2, 9.896161**2]),
        'SAPPHIRE': make_sellmeier_glass(1, [1.4313493, 0.65054713, 5.3414021], [0.0726631**2, 0.1193242**2, 18.028251**2]),
        'CAF2': make_sellmeier_glass(1.33973, [0.69913, 0.11994, 4.35181], [0.09374**2, 21.18**2, 38.46**2]),
        'N-PSK53': make_sellmeier_glass(1.0, [1.3434087, 0.241417935, 0.952896897], [0.00675074317, 0.0219910513, 103.551457]),
        'VACUUM': lambda wavelengths: np.ones_like(wavelengths)
    }

    catalogues = ['schott_glass_catalogue_2018_09.csv', 'ohara_glass_catalogue_2019_08.csv']

    for fname in catalogues:
        catalogue = files('hcipy.optics').joinpath(fname)
        with catalogue.open() as f:
            _glass_catalogue.update(_parse_sellmeier_glass_catalogue(f))

def _parse_sellmeier_glass_catalogue(catalogue_file):
    data = np.loadtxt(catalogue_file, dtype=str, delimiter=',', skiprows=1)

    database = {}
    for di in data:
        database[di[0].replace(" ", "")] = make_sellmeier_glass(1, di[1:4].astype('float'), di[4::].astype('float'))

    return database



from hcipy import *

def test_config():
    Configuration()['test_config'] = 1
    assert Configuration()['test_config'] == 1

    Configuration().test_config = 2
    assert Configuration()['test_config'] == 2

    Configuration().update({'test_section': {'test_value': 3}})
    assert Configuration().test_section.test_value == 3



from hcipy import *
import numpy as np
import scipy

def test_zernike_modes():
    grid = make_pupil_grid(256)
    aperture_mask = make_circular_aperture(1)(grid) > 0

    modes = make_zernike_basis(200, 1, grid)

    assert np.abs(np.std(modes[0][aperture_mask])) < 2e-2

    for m in modes[1:]:
        assert np.abs(np.std(m[aperture_mask]) - 1) < 2e-2

    for i, m in enumerate(modes):
        zn, zm = noll_to_zernike(i + 1)
        assert np.allclose(m, zernike(zn, zm, grid=grid))

def test_zernike_indices():
    for i in range(1, 500):
        n, m = noll_to_zernike(i)
        assert i == zernike_to_noll(n, m)

        n, m = ansi_to_zernike(i)
        assert i == zernike_to_ansi(n, m)

def test_disk_harmonic_modes():
    grid = make_pupil_grid(128)
    aperture_mask = make_circular_aperture(1)(grid) > 0

    num_modes = 20

    for bc in ['dirichlet', 'neumann']:
        modes = make_disk_harmonic_basis(grid, num_modes, bc=bc)

        for i, m1 in enumerate(modes):
            for j, m2 in enumerate(modes):
                product = np.sum((m1 * m2)[aperture_mask])
                assert np.abs(product - np.eye(num_modes)[i, j]) < 1e-2

def test_lp_modes():
    grid = make_pupil_grid(128)

    # Test for single-mode
    modes = make_lp_modes(grid, 2.4, 0.1, return_betas=False)
    assert len(modes) == 1

    # Test orthogonality
    modes = make_lp_modes(grid, 25, 0.1, return_betas=False)
    for i, m1 in enumerate(modes):
        for j, m2 in enumerate(modes):
            product = np.real(np.sum(m1 * m2.conj() * grid.weights))
            assert np.abs(product - 1 if i == j else 0) <= 1e-2

def test_sparse_mode_basis():
    transformation_matrix = np.empty((100 * 100, 100))
    for i in range(100):
        transformation_matrix[:, i] = np.random.randn(100 * 100)
        transformation_matrix[np.random.choice(100 * 100, 80 * 100, False), i] = 0

    mode_basis_1 = ModeBasis(transformation_matrix)

    assert mode_basis_1.is_dense
    assert not mode_basis_1.is_sparse
    assert not scipy.sparse.issparse(mode_basis_1.transformation_matrix)

    mode_basis_2 = mode_basis_1.to_dense(copy=False)

    assert mode_basis_2 is mode_basis_1

    mode_basis_3 = mode_basis_1.to_dense(copy=True)

    assert mode_basis_3 is not mode_basis_1
    assert np.allclose(mode_basis_1.transformation_matrix, mode_basis_3.transformation_matrix)

    mode_basis_4 = mode_basis_1.to_sparse()

    assert mode_basis_4.is_sparse
    assert not mode_basis_4.is_dense
    assert scipy.sparse.issparse(mode_basis_4.transformation_matrix)
    assert mode_basis_4.transformation_matrix.nnz < np.prod(mode_basis_4.transformation_matrix.shape)

    mode_basis_5 = mode_basis_4.to_dense()

    assert mode_basis_5.is_dense
    assert not mode_basis_5.is_sparse
    assert np.allclose(mode_basis_1.transformation_matrix, mode_basis_5.transformation_matrix)
    assert not scipy.sparse.issparse(mode_basis_5.transformation_matrix)

def test_gaussian_laguerre_modes():
    grid = make_focal_grid(32, 4)

    p_max = 5
    l_max = 5
    mode_field_diameter = 1

    modes = make_gaussian_laguerre_basis(grid, p_max, l_max, mode_field_diameter)
    num_modes = len(modes)

    for i, m1 in enumerate(modes):
        for j, m2 in enumerate(modes):
            product = np.sum((m1.conj() * m2) * grid.weights).real
            assert np.abs(product - np.eye(num_modes)[i, j]) < 1e-6

def test_gaussian_hermite_modes():
    grid = make_focal_grid(32, 4)

    num_modes = 50
    mode_field_diameter = 1

    modes = make_gaussian_hermite_basis(grid, num_modes, mode_field_diameter)

    for i, m1 in enumerate(modes):
        for j, m2 in enumerate(modes):
            product = np.sum((m1.conj() * m2) * grid.weights).real
            assert np.abs(product - np.eye(num_modes)[i, j]) < 1e-6

def test_fourier_modes():
    grid = make_pupil_grid(32)
    fourier_grid = make_fft_grid(grid, 1, 0.2)

    # Cosine modes
    cosine_modes = make_cosine_basis(grid, fourier_grid)
    num_modes = len(cosine_modes)

    for i, m1 in enumerate(cosine_modes):
        for j, m2 in enumerate(cosine_modes):
            product = np.sum((m1.conj() * m2) * grid.weights).real
            assert np.abs(product - np.eye(num_modes)[i, j]) < 1e-12

    # Sine modes
    sine_modes = make_sine_basis(grid, fourier_grid)
    num_modes = len(sine_modes)

    for i, m1 in enumerate(sine_modes):
        for j, m2 in enumerate(sine_modes):
            product = np.sum((m1.conj() * m2) * grid.weights).real
            assert np.abs(product - np.eye(num_modes)[i, j]) < 1e-12

    # Fourier basis
    fourier_modes = make_fourier_basis(grid, fourier_grid)
    num_modes = len(fourier_modes)

    for i, m1 in enumerate(fourier_modes):
        for j, m2 in enumerate(fourier_modes):
            product = np.sum((m1.conj() * m2) * grid.weights).real
            assert np.abs(product - np.eye(num_modes)[i, j]) < 1e-12

    # Complex Fourier basis
    complex_fourier_modes = make_complex_fourier_basis(grid, fourier_grid)
    num_modes = len(complex_fourier_modes)

    for i, m1 in enumerate(complex_fourier_modes):
        for j, m2 in enumerate(complex_fourier_modes):
            product = np.sum((m1.conj() * m2) * grid.weights).real
            assert np.abs(product - np.eye(num_modes)[i, j]) < 1e-12



from __future__ import division
import functools
import inspect

import numpy as np
from matplotlib.path import Path

from ..field import Field, make_hexagonal_grid
from ..dev import deprecated_name_changed

def make_circular_aperture(diameter, center=None):
    '''Makes a Field generator for a circular aperture.

    Parameters
    ----------
    diameter : scalar
        The diameter of the aperture.
    center : array_like
        The center of the aperture

    Returns
    -------
    Field generator
        This function can be evaluated on a grid to get a Field.
    '''
    if center is None:
        shift = np.zeros(2)
    else:
        shift = center * np.ones(2)

    def func(grid):
        if grid.is_('cartesian'):
            if grid.is_separated:
                x, y = grid.separated_coords
                x = x[np.newaxis, :]
                y = y[:, np.newaxis]
            else:
                x, y = grid.coords

            f = (((x - shift[0])**2 + (y - shift[1])**2) <= (diameter / 2)**2).ravel()
        else:
            f = grid.as_('polar').r <= (diameter / 2)

        return Field(f.astype('float'), grid)

    return func

def make_elliptical_aperture(diameters, center=None, angle=0):
    '''Makes a Field generator for an elliptical aperture.

    Parameters
    ----------
    diameters : array_like
        The diameters of the aperture in the two directions.
    center : array_like or None
        The center of the aperture.
    angle : scalar
        The orientation of the ellipse in radians.

    -------
    Field generator
        This function can be evaluated on a grid to get a Field.
    '''
    if center is None:
        shift = np.zeros(2)
    else:
        shift = center * np.ones(2)

    # Pre-calculate the minor and major axes dimensions after rotation
    major_axis = diameters[0] / 2
    minor_axis = diameters[1] / 2

    cos_angle_major = np.cos(angle) / major_axis
    cos_angle_minor = np.cos(angle) / minor_axis

    sin_angle_major = np.sin(angle) / major_axis
    sin_angle_minor = np.sin(angle) / minor_axis

    def func(grid):
        g = grid.as_('cartesian').shifted(shift)

        if g.is_separated:
            x, y = g.separated_coords
            x = x[np.newaxis, :]
            y = y[:, np.newaxis]
        else:
            x, y = g.coords

        term1 = ((x * cos_angle_major) - (y * sin_angle_major))**2
        term2 = ((x * sin_angle_minor) + (y *  cos_angle_minor))**2
        f = (term1 + term2) <= 1

        return Field(f.ravel().astype('float'), grid)

    return func

def make_rectangular_aperture(size, center=None):
    '''Makes a Field generator for a rectangular aperture.

    Parameters
    ----------
    size : scalar or array_like
        The length of the sides. If this is scalar, a square aperture is assumed.
    center : array_like
        The center of the aperture

    Returns
    -------
    Field generator
        This function can be evaluated on a grid to get a Field.
    '''
    half_dim = size * np.ones(2) / 2

    if center is None:
        shift = np.zeros(2)
    else:
        shift = center * np.ones(2)

    def func(grid):
        g = grid.as_('cartesian')

        if g.is_separated:
            x, y = g.separated_coords
            x = x[np.newaxis, :]
            y = y[:, np.newaxis]
        else:
            x, y = g.coords

        f = (np.abs(x - shift[0]) <= half_dim[0]) * (np.abs(y - shift[1]) <= half_dim[1])

        return Field(f.ravel().astype('float'), grid)

    return func

def make_irregular_polygon_aperture(vertices):
    '''Make an irregular polygonal aperture.

    Parameters
    ----------
    vertices : iterable of length 2 iterables
        The vertices of the polygon.

    Returns
    -------
    Field generator
        The generator for the polygonal aperture.
    '''
    p = Path(vertices)

    bounding_box_min = np.min(vertices, axis=0)
    bounding_box_max = np.max(vertices, axis=0)

    size = bounding_box_max - bounding_box_min
    center = (bounding_box_min + bounding_box_max) / 2

    def func(grid):
        res = grid.zeros()

        mask = make_rectangular_aperture(size, center=center)(grid).astype('bool')
        res[mask] = p.contains_points(grid.points[mask])

        return res

    return func

def make_regular_polygon_aperture(num_sides, circum_diameter, angle=0, center=None):
    '''Makes a Field generator for a regular-polygon-shaped aperture.

    Parameters
    ----------
    num_sides : integer
        The number of sides for the polygon.
    circum_diameter : scalar
        The circumdiameter of the polygon.
    angle : scalar
        The angle by which to turn the polygon.
    center : array_like
        The center of the aperture

    Returns
    -------
    Field generator
        This function can be evaluated on a grid to get a Field.
    '''
    if num_sides < 3:
        raise ValueError('The number of sides for a regular polygon has to greater or equal to 3.')

    if center is None:
        shift = np.zeros(2)
    else:
        shift = center * np.ones(2)

    epsilon = 1e-6

    apothem = np.cos(np.pi / num_sides) * circum_diameter / 2
    apothem += apothem * epsilon

    if center is None:
        shift = np.zeros(2)
    else:
        shift = center * np.ones(2)

    # Make use of symmetry
    if num_sides % 2 == 0:
        thetas = np.arange(int(num_sides / 2), dtype='float') * np.pi / int(num_sides / 2) + angle
    else:
        thetas = np.arange(int(num_sides / 2) + 1) * (num_sides - 2) * np.pi / (num_sides / 2) + angle

    mask = make_rectangular_aperture(circum_diameter)

    def func(grid, return_with_mask=False):
        g = grid.as_('cartesian')

        if g.is_separated:
            x, y = g.separated_coords

            x = x - shift[0]
            y = y - shift[1]

            ind_x = np.flatnonzero(x**2 < ((circum_diameter / 2)**2))
            if not len(ind_x):
                if return_with_mask:
                    return np.array([]), (slice(0, 0), slice(0, 0))
                else:
                    return grid.zeros()

            ind_y = np.flatnonzero(y**2 < ((circum_diameter / 2)**2))
            if not len(ind_y):
                if return_with_mask:
                    return np.array([]), (slice(0, 0), slice(0, 0))
                else:
                    return grid.zeros()

            m_x = slice(ind_x[0], ind_x[-1] + 1)
            m_y = slice(ind_y[0], ind_y[-1] + 1)

            x = x[m_x]
            y = y[m_y]

            f_sub = np.ones((len(ind_y), len(ind_x)))

            if num_sides % 2 == 0:
                for theta in thetas:
                    f_sub *= ((np.cos(theta) * x)[np.newaxis, :] + (np.sin(theta) * y)[:, np.newaxis])**2 <= apothem**2
            else:
                for theta in thetas:
                    f_sub *= (np.abs(np.sin(theta) * x)[np.newaxis, :] - (np.cos(theta) * y)[:, np.newaxis]) <= apothem

            if return_with_mask:
                return f_sub, (m_y, m_x)

            f = np.zeros(g.shape)
            f[m_y, m_x] = f_sub

            return Field(f.ravel(), grid)

        # Slow backup method
        m = mask(g) != 0

        x, y = g.coords
        x = x[m] - shift[0]
        y = y[m] - shift[1]

        f_sub = np.ones(x.size, dtype='float')

        # Make use of symmetry
        if num_sides % 2 == 0:
            for theta in thetas:
                f_sub *= (np.cos(theta) * x + np.sin(theta) * y)**2 <= apothem**2
        else:
            for theta in thetas:
                f_sub *= (np.abs(np.sin(theta) * x) + -np.cos(theta) * y) <= apothem

        if return_with_mask:
            return f_sub, m

        f = grid.zeros()
        f[m] = f_sub

        return Field(f, grid)

    return func

# Convenience function
def make_hexagonal_aperture(circum_diameter, angle=0, center=None):
    '''Makes a Field generator for a hexagon aperture.

    Parameters
    ----------
    circum_diameter : scalar
        The circumdiameter of the polygon.
    angle : scalar
        The angle by which to turn the hexagon.
    center : array_like
        The center of the aperture

    Returns
    -------
    Field generator
        This function can be evaluated on a grid to get a Field.
    '''
    return make_regular_polygon_aperture(6, circum_diameter, angle, center)

def make_spider(p1, p2, spider_width):
    '''Make a rectangular obstruction from `p1` to `p2`.

    Parameters
    ----------
    p1 : list or ndarray
        The starting coordinates of the spider.
    p2 : list or ndarray
        The end coordinates of the spider.
    spider_width : scalar
        The full width of the spider.

    Returns
    -------
    Field generator
        The spider obstruction.
    '''
    delta = np.array(p2) - np.array(p1)
    shift = delta / 2 + np.array(p1)

    spider_angle = np.arctan2(delta[1], delta[0])
    spider_length = np.linalg.norm(delta)

    def func(grid):
        g = grid.as_('cartesian')

        if g.is_separated:
            x, y = g.separated_coords
            x = x[np.newaxis, :]
            y = y[:, np.newaxis]
        else:
            x, y = g.coords

        x = x - shift[0]
        y = y - shift[1]

        x_new = x * np.cos(spider_angle) + y * np.sin(spider_angle)
        y_new = y * np.cos(spider_angle) - x * np.sin(spider_angle)

        # Doing comparisons for each side separately is actually
        # faster than doing abs(x_new) < (spider_length / 2).
        spider = x_new <= (spider_length / 2)
        spider *= x_new >= (-spider_length / 2)
        spider *= y_new <= (spider_width / 2)
        spider *= y_new >= (-spider_width / 2)

        return Field(1 - spider.ravel(), grid)
    return func

def make_spider_infinite(p, angle, spider_width):
    '''Make an infinite spider starting at `p` and extending at an angle `angle`.

    Parameters
    ----------
    p : list or ndarray
        The starting coordinate of the spider.
    angle : scalar
        The angle to which the spider is pointing in degrees.
    spider_width : scalar
        The full width of the spider.

    Returns
    -------
    Field generator
        The spider obstruction.
    '''
    spider_angle = np.radians(angle)

    def func(grid):
        g = grid.as_('cartesian')

        if g.is_separated:
            x, y = g.separated_coords
            x = x[np.newaxis, :]
            y = y[:, np.newaxis]
        else:
            x, y = g.coords

        x = x + p[0]
        y = y + p[1]

        x_new = x * np.cos(spider_angle) + y * np.sin(spider_angle)
        y_new = y * np.cos(spider_angle) - x * np.sin(spider_angle)

        # Doing comparisons for each side separately is actually
        # faster than doing abs(y_new) < (spider_length / 2).
        infinite_spider = y_new <= (spider_width / 2)
        infinite_spider *= y_new >= (-spider_width / 2)
        infinite_spider *= x_new >= 0

        return Field(1 - infinite_spider.ravel(), grid)
    return func

def make_obstructed_circular_aperture(pupil_diameter, central_obscuration_ratio, num_spiders=0, spider_width=0.01):
    '''Make a simple circular aperture with central obscuration and support structure.

    Parameters
    ----------
    pupil_diameter : scalar
        The diameter of the circular aperture.
    central_obscuration_ratio : scalar
        The ratio of the diameter of the central obscuration compared to the pupil diameter.
    num_spiders : int
        The number of spiders holding up the central obscuration.
    spider_width : scalar
        The full width of the spiders.

    Returns
    -------
    Field generator
        The circularly obstructed aperture.
    '''
    central_obscuration_diameter = pupil_diameter * central_obscuration_ratio

    def func(grid):
        pupil_outer = make_circular_aperture(pupil_diameter)(grid)
        pupil_inner = make_circular_aperture(central_obscuration_diameter)(grid)
        spiders = 1

        spider_angles = np.linspace(0, 2 * np.pi, num_spiders, endpoint=False)

        for angle in spider_angles:
            x = pupil_diameter * np.cos(angle)
            y = pupil_diameter * np.sin(angle)

            spiders *= make_spider((0, 0), (x, y), spider_width)(grid)

        return (pupil_outer - pupil_inner) * spiders
    return func

def make_obstruction(aperture):
    '''Create an obstruction of `aperture`.

    Parameters
    ----------
    aperture : Field generator
        The aperture to invert.

    Returns
    -------
    Field generator
        The obstruction.
    '''
    return lambda grid: 1 - aperture(grid)

def make_rotated_aperture(aperture, angle):
    '''Create a rotated version of `aperture`.

    Parameters
    ----------
    aperture : Field generator
        The aperture to rotate.
    angle : scalar
        The angle in radians by which to rotate the aperture.

    Returns
    -------
    Field generator
        The rotated aperture.
    '''
    return lambda grid: Field(aperture(grid.rotated(-angle)), grid)

def make_shifted_aperture(aperture, shift):
    '''Create a shifted version of `aperture`.

    Parameters
    ----------
    aperture : Field generator
        The aperture to rotate.
    shift : array_like
        The shift of the aperture.

    Returns
    -------
    Field generator
        The shifted aperture.
    '''
    return lambda grid: Field(aperture(grid.shifted(-np.array(shift))), grid)

def make_segmented_aperture(segment_shape, segment_positions, segment_transmissions=1, return_segments=False):
    '''Create a segmented aperture.

    Parameters
    ----------
    segment_shape : Field generator
        The shape for each of the segments.
    segment_positions : Grid
        The center position for each of the segments.
    segment_transmissions : scalar or ndarray
        The transmission for each of the segments. If this is a scalar, the same transmission is used for all segments.
    return_segments : boolean
        Whether to return a ModeBasis of all segments as well.

    Returns
    -------
    Field generator
        The segmented aperture.
    list of Field generators
        The segments. Only returned if return_segments is True.
    '''
    segment_transmissions = np.ones(segment_positions.size) * segment_transmissions

    mask_available = 'return_with_mask' in inspect.signature(segment_shape).parameters

    def func(grid):
        res = grid.zeros(dtype=segment_transmissions.dtype)

        for p, t in zip(segment_positions.points, segment_transmissions):
            if mask_available:
                # Use the masked version of the segment shape.
                segment_sub, mask = segment_shape(grid.shifted(-p), return_with_mask=True)

                if isinstance(mask, tuple):
                    res.shaped[mask][segment_sub > 0.5] = t
                else:
                    res[mask][segment_sub] = t
            else:
                segment = segment_shape(grid.shifted(-p))
                res[segment > 0.5] = t

        return Field(res, grid)

    if return_segments:
        def seg(grid, p, t):
            return segment_shape(grid.shifted(-p)) * t

        segments = []
        for p, t in zip(segment_positions.points, segment_transmissions):
            segments.append(functools.partial(seg, p=p, t=t))

        return func, segments
    else:
        return func

def make_hexagonal_segmented_aperture(num_rings, segment_flat_to_flat, gap_size, starting_ring=0, return_segments=False):
    '''Create a hexagonal segmented aperture.

    All segments have a flat-to-flat size of `segment_flat_to_flat`, which means that
    the segment pitch (the distance between the centers of adjacent segments) is the sum
    of the segment flat-to-flat size and the gap size.

    Parameters
    ----------
    num_rings : int
        The number of rings of hexagons to include, not counting the central segment.
        So 2 for a JWST-like aperture, 3 for a Keck-like aperture, and so on.
    segment_flat_to_flat : scalar
        The distance between sides (flat-to-flat) of a single segment.
    gap_size : scalar
        The gap between adjacent segments.
    starting_ring : int
        The first ring of segments. This can be used to exclude the center segment (by
        setting it to one), or the center segment and first ring (by setting it to two).
        The default (zero) includes the center segment.
    return_segments : boolean
        Whether to return a ModeBasis of all segments as well.

    Returns
    -------
    Field generator
        The segmented aperture.
    list of Field generators
        The segments. Only returned if return_segments is True.
    '''
    segment_circum_diameter = segment_flat_to_flat * 2 / np.sqrt(3)
    segment = make_hexagonal_aperture(segment_circum_diameter, np.pi / 2)

    segment_pitch = segment_flat_to_flat + gap_size
    segment_positions = make_hexagonal_grid(segment_pitch, num_rings, pointy_top=False)

    if starting_ring != 0:
        starting_segment = 3 * (starting_ring - 1) * starting_ring + 1

        mask = segment_positions.zeros(dtype='bool')
        mask[starting_segment:] = True

        segment_positions = segment_positions.subset(mask)

    return make_segmented_aperture(segment, segment_positions, return_segments=return_segments)

@deprecated_name_changed(make_circular_aperture)
def circular_aperture():
    pass

@deprecated_name_changed(make_elliptical_aperture)
def elliptical_aperture():
    pass

@deprecated_name_changed(make_rectangular_aperture)
def rectangular_aperture():
    pass

@deprecated_name_changed(make_hexagonal_aperture)
def hexagonal_aperture():
    pass

@deprecated_name_changed(make_regular_polygon_aperture)
def regular_polygon_aperture():
    pass

@deprecated_name_changed(make_irregular_polygon_aperture)
def irregular_polygon_aperture():
    pass



from .atmospheric_model import Cn_squared_from_fried_parameter
from .infinite_atmospheric_layer import InfiniteAtmosphericLayer

import numpy as np

def make_standard_atmospheric_layers(input_grid, L0=10):
    heights = np.array([500, 1000, 2000, 4000, 8000, 16000])
    velocities = np.array([10, 10, 10, 10, 10, 10])
    Cn_squared = np.array([0.2283, 0.0883, 0.0666, 0.1458, 0.3350, 0.1350]) * 1e-12

    layers = []
    for h, v, cn in zip(heights, velocities, Cn_squared):
        layers.append(InfiniteAtmosphericLayer(input_grid, cn, L0, v, h, 2))

    return layers

def make_las_campanas_atmospheric_layers(input_grid, r0=0.16, L0=25, wavelength=550e-9):
    '''Creates a multi-layer atmosphere for the Las Campanas Observatory site.

    The layer parameters are taken from [Males2019]_ who based it on site testing from [Prieto2010]_ and [Osip2011]_ .

    .. [Prieto2010] G. Prieto et al., “Giant Magellan telescope site testing seeing and
        turbulence statistics,” Proc. SPIE 7733, 77334O (2010).

    .. [Osip2011] Joanna E. Thomas-Osip et al. "Giant Magellan Telescope Site Testing Summary."
        arXiv:1101.2340 (2011).

    .. [Males2019] Jared Males et al. "Ground-based adaptive optics coronagraphic performance
        under closed-loop predictive control", JATIS, Volume 4, id. 019001 (2018).

    Parameters
    ----------
    input_grid : Grid
        Th
    r0 : scalar
        The integrated Cn^2 value for the atmosphere.
    L0 : scalar
        The outer scale of the atmosphere
    wavelength : scalar
        The wavelength in meters at which to calculate the Fried parameter (default: 550nm).

    Returns
    -------
    list
        A list of turbulence layers.
    '''
    heights = np.array([250, 500, 1000, 2000, 4000, 8000, 16000])
    velocities = np.array([10, 10, 20, 20, 25, 30, 25])

    integrated_cn_squared = Cn_squared_from_fried_parameter(r0, wavelength=500e-9)
    Cn_squared = np.array([0.42, 0.03, 0.06, 0.16, 0.11, 0.10, 0.12]) * integrated_cn_squared

    layers = []
    for h, v, cn in zip(heights, velocities, Cn_squared):
        layers.append(InfiniteAtmosphericLayer(input_grid, cn, L0, v, h, 2))

    return layers

def make_keck_atmospheric_layers(input_grid):
    '''Creates a multi-layer atmosphere for Keck Observatory.

    The atmospheric parameters are based off of [Keck AO note 303]_.

    .. [Keck AO note 303] https://www2.keck.hawaii.edu/optics/kpao/files/KAON/KAON303.pdf

    Parameters
    ----------
    input_grid : Grid
        The input grid of the atmospheric layers.

    Returns
    -------
    list
        A list of turbulence layers.
    '''
    heights = np.array([0, 2100, 4100, 6500, 9000, 12000, 14800])
    velocities = np.array([6.7, 13.9, 20.8, 29.0, 29.0, 29.0, 29.0])
    outer_scales = np.array([20, 20, 20, 20, 20, 20, 20])
    Cn_squared = np.array([0.369, 0.219, 0.127, 0.101, 0.046, 0.111, 0.027]) * 1e-12

    layers = []
    for h, v, cn, L0 in zip(heights, velocities, Cn_squared, outer_scales):
        layers.append(InfiniteAtmosphericLayer(input_grid, cn, L0, v, h, 2))

    return layers



import numpy as np
from math import sqrt
from scipy.special import eval_genlaguerre

def gaussian_laguerre(p, l, mode_field_diameter=1, grid=None):
    r'''Creates a Gaussian-Laguerre mode.

    This function evaluates a (p,l) order Gaussian-Laguerre mode on a grid.
    The definition of the modes are the following,

    .. math::
        \exp{\left(-\frac{r^2}{w_0^2}\right)} L_p^{|l|}\left(\frac{2r^2}{w_0^2} \right) \left(\sqrt{2}\frac{r}{w_0}\right)

    Here :math:`w_0` is the mode_field_radius, which is :math:`\mathrm{MFD}/2`.
    And :math:`L_p^{|l|}` are the generalized Laguerre Polynomials.
    All modes are numerical normalized to have a total power of 1.

    More details on generalized Laguerre Polynomials can be found on:
    http://mathworld.wolfram.com/AssociatedLaguerrePolynomial.html

    Parameters
    ----------
    p : int
        The radial order.
    l : int
        The azimuthal order.
    mode_field_diameter : scalar
        The mode field diameter of the mode.
    grid : Grid
        The grid on which to evaluate the mode.

    Returns
    -------
    Field
        The evaluated mode.
    '''
    from ..field import Field

    if grid.is_separated and grid.is_('polar'):
        R, Theta = grid.separated_coords
    else:
        R, Theta = grid.as_('polar').coords

    # Easy access
    r = 2 * R / mode_field_diameter
    r2 = r**2

    # Compute the mode
    lg = (r * sqrt(2))**abs(l) * np.exp(-r2) * np.exp(-1j * l * Theta) * eval_genlaguerre(p, abs(l), 2 * r2)

    # Numerically normalize the mode
    lg /= np.sqrt(np.sum(np.abs(lg)**2 * grid.weights))

    return Field(lg, grid)

# High level functions
def make_gaussian_laguerre_basis(grid, pmax, lmax, mode_field_diameter, pmin=0):
    '''Creates a Gaussian-Laguerre mode basis.

    This function evaluates Gaussian-Laguerre modes. For each radial order
    within [pmin, pmax] it will calculate the azimuthal order [-lmax, lmax] inclusive.
    This function returns a ModeBasis made out of these Gaussian-Laguerre modes.

    Parameters
    ----------
    grid : Grid
        The grid on which to evaluate the Gaussian-Laguerre mode.
    pmax : int
        The maximum radial order of the modes.
    lmax : int
        The maximum azimuthal order.
    mode_field_diameter : scalar
        The mode field diameter of the Gaussian-Laguerre mode.
    pmin : int
        The minimal radial order.

    Returns
    -------
    ModeBasis
        The Gaussian-Laguerre modes.
    '''
    from .mode_basis import ModeBasis

    modes = [gaussian_laguerre(pi, li, mode_field_diameter, grid) for li in range(-lmax, lmax + 1) for pi in range(pmin, pmax)]
    return ModeBasis(modes, grid)



from .wavefront_sensor import WavefrontSensorOptics, WavefrontSensorEstimator
from ..propagation import FraunhoferPropagator
from ..optics import PhaseApodizer, LinearRetarder, Wavefront
from ..aperture import make_circular_aperture
from ..field import make_uniform_grid, Field

import numpy as np

class ZernikeWavefrontSensorOptics(WavefrontSensorOptics):
    '''The optical elements for a Zernike wavefront sensor.

    This class uses a propagation scheme of [NDiaye2013]_, similar to that of Lyot coronagraphs
    with a small blocking mask, see [Soummer2007]_.

    .. [NDiaye2013] N'Diaye et al. 2013, "Calibration of quasi-static
        aberrations in exoplanet direct-imaging instruments with a
        Zernike phase-mask sensor"

    .. [Soummer2007] Soummer et al. 2007, "Fast computation of Lyot-style
        coronagraph propagation"

    Parameters
    ----------
    input_grid : Grid
        The grid on which the input wavefront is defined.
    phase_step : scalar
        The phase of the phase dot of the zernike wavefront sensor. The default is pi/2.
    phase_dot_diameter : scalar
        The diameter of the phase dot. This has units of lambda_0/D.
    num_pix : scalar
        The number of pixels across the phase dot.
    pupil_diameter : scalar
        The diameter of the pupil. This is used for calculating the size of the phase dot.
    reference_wavelength : scalar
        The reference wavelength. This is used for calcualting the size of the phase dot.
    '''
    def __init__(self, input_grid, phase_step=np.pi / 2, phase_dot_diameter=1.06, num_pix=128, pupil_diameter=1, reference_wavelength=1):
        self.input_grid = input_grid
        self.output_grid = input_grid

        # Make the phase dot
        phase_dot_diameter *= reference_wavelength / pupil_diameter
        focal_grid = make_uniform_grid([num_pix, num_pix], phase_dot_diameter)
        self.phase_dot = PhaseApodizer(make_circular_aperture(phase_dot_diameter)(focal_grid) * phase_step)

        # Make the propagator
        self.prop = FraunhoferPropagator(input_grid, focal_grid)

    def forward(self, wavefront):
        '''Propagates a wavefront through the wavefront sensor.

        Parameters
        ----------
        wavefront : Wavefront
            The input wavefront that will propagate through the system.

        Returns
        -------
        wf : Wavefront
            The output wavefront.
        '''
        wf_foc = self.prop.forward(wavefront)
        wf_foc.electric_field -= self.phase_dot.forward(wf_foc).electric_field

        pup = self.prop.backward(wf_foc)
        pup.electric_field[:] = wavefront.electric_field - pup.electric_field

        return pup

    def backward(self, wavefront):
        '''Propagates a wavefront backwards through the wavefront sensor.

        Parameters
        ----------
        wavefront : Wavefront
            The input wavefront that will propagate through the system.

        Returns
        -------
        wf : Wavefront
            The output wavefront.
        '''
        wf_foc = self.prop.forward(wavefront)
        wf_foc.electric_field -= self.phase_dot.backward(wf_foc).electric_field

        pup = self.prop.backward(wf_foc)
        pup.electric_field[:] = wavefront.electric_field - pup.electric_field

        return pup

class ZernikeWavefrontSensorEstimator(WavefrontSensorEstimator):
    '''Estimates the wavefront slopes from pyramid wavefront sensor images.

    Parameters
    ----------
    aperture : function
        A function which mask the output phase measurements.
    output_grid : Grid
        The grid on which the output of a pyramid wavefront sensor is sampled.
    reference : Field
        A reference image to subtract from the Zernike wavefront sensor data.

    Attributes
    ----------
    measurement_grid : Grid
        The grid on which the phase measurements are defined.
    pupil_mask : array_like
        A mask for the phase measurements.
    num_measurements : int
        The number of pixels in the output vector.
    reference : Field
        A reference image to subtract from the Zernike wavefront sensor data.
    '''
    def __init__(self, aperture, output_grid, reference):
        self.measurement_grid = output_grid
        self.pupil_mask = aperture(self.measurement_grid)
        self.reference = reference
        self.num_measurements = int(np.sum(self.pupil_mask > 0))

    def estimate(self, images):
        '''A function which estimates the phase from a Zernike wavefront sensor image.

        Parameters
        ----------
        images - List
            A list of scalar intensity fields containing Zernike wavefront sensor images.

        Returns
        -------
        res - Field
            A field with phase estimates.
        '''
        image = images[0]

        intensity_measurements = (image - self.reference).ravel() * self.pupil_mask
        res = Field(intensity_measurements[self.pupil_mask > 0], self.pupil_mask.grid)
        return res

class VectorZernikeWavefrontSensorOptics(WavefrontSensorOptics):
    '''The optical elements for a vector-Zernike wavefront sensor.

    The vector Zernike is based on [Doelman2019]_.

    This class uses a propagation scheme of [NDiaye2013]_, similar to that of Lyot coronagraphs
    with a small blocking mask, see [Soummer2007]_. The reference wavefront is also propagated
    through a half-wave retarder because the vector-Zernike mask is a half-wave
    retarder.

    .. [Doelman2019] Doelman et al. 2019, "Simultaneous phase and amplitude aberration sensing
        with a liquid-crystal vector-Zernike phase mask"

    .. [NDiaye2013] N'Diaye et al. 2013, "Calibration of quasi-static
        aberrations in exoplanet direct-imaging instruments with a
        Zernike phase-mask sensor"

    .. [Soummer2007] Soummer et al. 2007, "Fast computation of Lyot-style
        coronagraph propagation"

    Parameters
    ----------
    input_grid : Grid
        The grid on which the input wavefront is defined.
    phase_retardation : scalar or Field
        The relative phase retardation induced between the fast and slow axis.
    phase_step : scalar
        The phase of the phase dot of the zernike wavefront sensor. The default is pi/2.
    phase_dot_diameter : scalar
        The diameter of the phase dot. This has units of lambda_0/D.
    num_pix : scalar
        The number of pixels across the phase dot.
    pupil_diameter : scalar
        The diameter of the pupil. This is used for calculating the size of the phase dot.
    reference_wavelength : scalar
        The reference wavelength. This is used for calcualting the size of the phase dot.
    '''
    def __init__(self, input_grid, phase_retardation=np.pi, phase_step=np.pi / 2, phase_dot_diameter=1.06, num_pix=128, pupil_diameter=1, reference_wavelength=1):
        self.input_grid = input_grid
        self.output_grid = input_grid
        self._phase_retardation = phase_retardation

        # Make the vector-Zernike wavefront sensor mask
        phase_dot_diameter *= reference_wavelength / pupil_diameter
        focal_grid = make_uniform_grid([num_pix, num_pix], phase_dot_diameter)
        phase_dot = make_circular_aperture(phase_dot_diameter)(focal_grid) * phase_step

        self.vZWFS_mask = LinearRetarder(phase_retardation, phase_dot / 2)

        # Make half-wave plate for reference offset
        self.HWP = LinearRetarder(phase_retardation, 0)

        # Make the propagator
        self.prop = FraunhoferPropagator(input_grid, focal_grid)

    @property
    def phase_retardation(self):
        '''The phase retardation of the linear retarder
        '''
        return self._phase_retardation

    @phase_retardation.setter
    def phase_retardation(self, phase_retardation):
        self._phase_retardation = phase_retardation

        self.vZWFS_mask.phase_retardation = phase_retardation
        self.HWP.phase_retardation = phase_retardation

    def forward(self, wavefront):
        '''Propagates a wavefront through the wavefront sensor.

        Parameters
        ----------
        wavefront : Wavefront
            The input wavefront that will propagate through the system.

        Returns
        -------
        wf : Wavefront
            The output wavefront.
        '''
        wf_foc = self.prop.forward(wavefront)
        if wf_foc.is_scalar:
            wf_foc = Wavefront(wf_foc.electric_field, wavelength=wf_foc.wavelength, input_stokes_vector=(1, 0, 0, 0))

        wf_foc.electric_field = self.HWP.forward(wf_foc).electric_field - self.vZWFS_mask.forward(wf_foc).electric_field

        pup = self.prop.backward(wf_foc)
        pup.electric_field[:] = self.HWP.forward(wavefront).electric_field - pup.electric_field

        return pup

    def backward(self, wavefront):
        '''Propagates a wavefront backwards through the wavefront sensor.

        Parameters
        ----------
        wavefront : Wavefront
            The input wavefront that will propagate through the system.

        Returns
        -------
        wf : Wavefront
            The output wavefront.
        '''
        wf_foc = self.prop.forward(wavefront)
        if wf_foc.is_scalar:
            wf_foc = Wavefront(wf_foc.electric_field, wavelength=wf_foc.wavelength, input_stokes_vector=(1, 0, 0, 0))

        wf_foc.electric_field = self.HWP.backward(wf_foc).electric_field - self.vZWFS_mask.backward(wf_foc).electric_field

        pup = self.prop.backward(wf_foc)
        pup.electric_field[:] = self.HWP.backward(wavefront).electric_field - pup.electric_field

        return pup
